import json
from typing import Any

from .request import Request


class OAI3JSON:
    requests: Any = None
    formatted: Any = None
    models: Any = None
    security_schemes: Any = None

    def __init__(self, oai3_schema_resolver):
        self.oai3_schema_resolver = oai3_schema_resolver
        self.requests = []
        self.formatted = []
        self.models = {}

    def set_requests(self, requests):
        self.requests = requests
        self.formatted = [self.format_request(request) for request in self.requests]

    def set_components(self, components):
        supported = ["models", "securitySchemes"]
        for key in components.keys():
            if key not in supported:
                raise ValueError(
                    f"Attempt to set unsupported OpenAPI3.0 component which is not currently supported: {key}"
                )
        if "models" in components:
            self.set_models(components["models"])
        if "securitySchemes" in components:
            self.set_security_schemes(components["securitySchemes"])

    def set_models(self, models):
        self.models = models

    def set_security_schemes(self, security_schemes):
        self.security_schemes = security_schemes

    def format_request(self, request):
        formatted_request = Request(self.oai3_schema_resolver)
        formatted_request.set_request(request)
        return formatted_request

    def pretty(self, root_properties=None):
        return self.as_string(pretty=True, root_properties=root_properties)

    def compact(self, root_properties=None):
        return self.as_string(pretty=False, root_properties=root_properties)

    def as_string(self, pretty=False, root_properties=None):
        data = self.convert()
        if root_properties is not None:
            data = {**data, **root_properties}
        if pretty:
            return json.dumps(data, indent=2, sort_keys=True)
        return json.dumps(data)

    def convert(self):
        paths: dict[str, Any] = {}
        for request in self.formatted:
            absolute_path = "/" + request.relative_path.lstrip("/")
            if absolute_path not in paths:
                paths[absolute_path] = {}

            path_data = request.convert()
            for request_method, path_doc in path_data.items():
                if request_method in paths[absolute_path]:
                    raise ValueError(f"Two routes had the same path and method: {absolute_path} - {request_method}")
                paths[absolute_path][request_method] = path_doc

        data: dict[str, Any] = {
            "openapi": "3.0.0",
            "paths": paths,
            "components": {},
        }

        if self.models:
            data["components"]["schemas"] = {
                model_name: self.oai3_schema_resolver(model).convert() for (model_name, model) in self.models.items()
            }

        if self.security_schemes:
            data["components"]["securitySchemes"] = {name: data for (name, data) in self.security_schemes.items()}

        return data



from typing import Any

from clearskies.configs import config


class ListAnyDict(config.Config):
    def __set__(self, instance, value: list[dict[str, Any]]):
        if not isinstance(value, list):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a parameter that requries a list."
            )
        for index, list_item in enumerate(value):
            if not isinstance(list_item, dict):
                error_prefix = self._error_prefix(instance)
                raise TypeError(
                    f"{error_prefix} I was expecting a list of dictionaries, but item # {index + 1} has type '{list_item.__class__.__name__}."
                )
            for key, val in list_item.items():
                if not isinstance(key, str):
                    error_prefix = self._error_prefix(instance)
                    raise TypeError(
                        f"{error_prefix} attempt to set a dictionary with a non-string key for item #{index + 1}."
                    )
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> list[dict[str, Any]]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Self, overload

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.columns.many_to_many_ids import ManyToManyIds

if TYPE_CHECKING:
    from clearskies import Model


class ManyToManyIdsWithData(ManyToManyIds):
    """
    A column to represent a many-to-many relationship with information stored in the relationship itself.

    This is an extention of the many-to-many column, but with one important addition: data about the
    relationship is stored in the pivot table.  This creates some differences, which are best
    explained by example:

    ```python
    import clearskies


    class ThingyWidgets(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        # these could also be belongs to relationships, but the pivot model
        # is rarely used directly, so I'm being lazy to avoid having to use
        # model references.
        thingy_id = clearskies.columns.String()
        widget_id = clearskies.columns.String()
        name = clearskies.columns.String()
        kind = clearskies.columns.String()


    class Thingy(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()


    class Widget(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        thingy_ids = clearskies.columns.ManyToManyIdsWithData(
            related_model_class=Thingy,
            pivot_model_class=ThingyWidgets,
            readable_pivot_column_names=["id", "thingy_id", "widget_id", "name", "kind"],
        )
        thingies = clearskies.columns.ManyToManyModels("thingy_ids")
        thingy_widgets = clearskies.columns.ManyToManyPivots("thingy_ids")


    def my_application(widgets: Widget, thingies: Thingy):
        thing_1 = thingies.create({"name": "Thing 1"})
        thing_2 = thingies.create({"name": "Thing 2"})
        thing_3 = thingies.create({"name": "Thing 3"})
        widget = widgets.create({
            "name": "Widget 1",
            "thingy_ids": [
                {"thingy_id": thing_1.id, "name": "Widget Thing 1", "kind": "Special"},
                {"thingy_id": thing_2.id, "name": "Widget Thing 2", "kind": "Also Special"},
            ],
        })

        return widget


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            my_application,
            model_class=Widget,
            return_records=True,
            readable_column_names=["id", "name", "thingy_widgets"],
        ),
        classes=[Widget, Thingy, ThingyWidgets],
    )

    if __name__ == "__main__":
        cli()
    ```

    As with setting ids in the ManyToManyIds class, any items left out will result in the relationship
    (including all its related data) being removed.  An important difference with the ManyToManyWithData
    column is the way you specify which record is being connected.  This is easy for the ManyToManyIds column
    because all you provide is the id from the related model.  When working with the ManyToManyWithData
    column, you provide a dictionary for each relationship (so you can provide the data that goes in the
    pivot model).  To let it know what record is being connected, you therefore explicitly provide
    the id from the related model in a dictionary key with the name of the related model id column in
    the pivot (e.g. `{"thingy_id": id}` in the first example.  However, if there are unique columns in the
    related model, you can provide those instead.  If you execute the above example you'll get:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "c4be91a8-85a1-4e29-994a-327f59e26ec7",
            "name": "Widget 1",
            "thingy_widgets": [
                {
                    "id": "3a8f6f14-9657-49d8-8844-0db3452525fe",
                    "thingy_id": "db292ebc-7b2b-4306-aced-8e6d073ec264",
                    "widget_id": "c4be91a8-85a1-4e29-994a-327f59e26ec7",
                    "name": "Widget Thing 1",
                    "kind": "Special",
                },
                {
                    "id": "480a0192-70d9-4363-a669-4a59f0b56730",
                    "thingy_id": "d469dbe9-556e-46f3-bc48-03f8cb8d8e44",
                    "widget_id": "c4be91a8-85a1-4e29-994a-327f59e26ec7",
                    "name": "Widget Thing 2",
                    "kind": "Also Special",
                },
            ],
        },
        "pagination": {},
        "input_errors": {},
    }
    ```
    """

    """ The list of columns in the pivot model that can be set when saving data from an endpoint. """
    setable_column_names = configs.WriteableModelColumns("pivot_model_class")

    """ The list of columns in the pivot model that will be included when returning records from an endpoint. """
    readable_pivot_column_names = configs.ReadableModelColumns("pivot_model_class")

    """
    Complicated, but probably should be false.

    Sometimes you have to provide data from the related model class in your save data so that
    clearskies can find the right record.  Normally, this lookup column is not persisted to the
    pivot table, because it is assumed to only exist in the related table.  In some cases though,
    you may want it in both, in which case you can set this to true.
    """
    persist_unique_lookup_column_to_pivot_table = configs.Boolean(default=False)

    default = configs.ListAnyDict(default=None)  #  type: ignore
    setable = configs.ListAnyDictOrCallable(default=None)  #  type: ignore
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        related_model_class,
        pivot_model_class,
        own_column_name_in_pivot: str = "",
        related_column_name_in_pivot: str = "",
        readable_related_columns: list[str] = [],
        readable_pivot_column_names: list[str] = [],
        setable_column_names: list[str] = [],
        persist_unique_lookup_column_to_pivot_table: bool = False,
        default: list[dict[str, Any]] = [],
        setable: list[dict[str, Any]] | Callable[..., list[dict[str, Any]]] = [],
        is_readable: bool = True,
        is_writeable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> list[Any]:
        pass

    def __get__(self, instance, cls):
        return super().__get__(instance, cls)

    def __set__(self, instance, value: list[dict[str, Any]]) -> None:  # type: ignore
        instance._next_data[self.name] = value

    def post_save(self, data, model, id):
        # if our incoming data is not in the data array or is None, then nothing has been set and we do not want
        # to make any changes
        if self.name not in data or data[self.name] is None:
            return data

        # figure out what ids need to be created or deleted from the pivot table.
        if not model:
            old_ids = set()
        else:
            old_ids = set(self.__get__(model, model.__class__))

        # this is trickier for many-to-many-with-data compared to many-to-many.  We're generally
        # expecting data[self.name] to be a list of dictionaries.  For each entry, we need to find
        # the corresponding entry in the pivot table to decide if we need to delete, create, or update.
        # However, since we have a dictionary there are a variety of ways that we can connect to
        # an entry in the related table - either related id or any unique column from the related
        # table.  Technically we might also specify a pivot id, but we're generally trying to be
        # transparent to those, so let's ignore that one.
        related_column_name_in_pivot = self.related_column_name_in_pivot
        own_column_name_in_pivot = self.own_column_name_in_pivot
        unique_related_columns = {
            column.name: column.name for column in self.related_columns.values() if column.is_unique
        }
        related_model = self.related_model
        pivot_model = self.pivot_model
        # minor cheating
        if hasattr(pivot_model.backend, "create_table"):
            pivot_model.backend.create_table(pivot_model)
        new_ids = set()
        for pivot_record in data[self.name]:
            # first we need to identify which related column this belongs to.
            related_column_id = None
            # if they provide the related column id in the pivot data then we're good
            if related_column_name_in_pivot in pivot_record:
                related_column_id = pivot_record[related_column_name_in_pivot]
            elif len(unique_related_columns):
                for pivot_column, pivot_value in pivot_record.items():
                    if pivot_column not in unique_related_columns:
                        continue
                    related = related_model.find(f"{pivot_column}={pivot_value}")
                    related_column_id = getattr(related, related.id_column_name)
                    if related_column_id:
                        # remove this column from the data - it was used to lookup the right
                        # record, but mostly won't exist in the model, unless we've been instructed
                        # to keep it
                        if not self._config.get("persist_unique_lookup_column_to_pivot_table"):  # type: ignore
                            del pivot_record[pivot_column]
                        break
            if not related_column_id:
                column_list = "'" + "', '".join(list(unique_related_columns.keys())) + "'"
                raise ValueError(
                    f"Missing data for {self.name}: Unable to match related record for a record in the many-to-many relationship: you must provide either '{related_column_name_in_pivot}' with the id column for the related table, or a value from one of the unique columns: {column_list}"
                )
            pivot = (
                pivot_model.where(f"{related_column_name_in_pivot}={related_column_id}")
                .where(f"{own_column_name_in_pivot}={id}")
                .first()
            )
            new_ids.add(related_column_id)
            # this will either update or create accordingly
            pivot.save(
                {
                    **pivot_record,
                    related_column_name_in_pivot: related_column_id,
                    own_column_name_in_pivot: id,
                }
            )

        # the above took care of isnerting and updating active records.  Now we need to delete
        # records that are no longer needed.
        to_delete = old_ids - new_ids
        if to_delete:
            for model_to_delete in pivot_model.where(
                f"{related_column_name_in_pivot} IN (" + ",".join(map(str, to_delete)) + ")"
            ):
                model_to_delete.delete()

        return data



from clearskies.configs import model_columns


class WriteableModelColumns(model_columns.ModelColumns):
    def get_allowed_columns(self, model_class, column_configs):
        return [name for (name, column) in column_configs.items() if column.is_writeable]

    def my_description(self):
        return "writeable column"



from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable

import clearskies.configs
import clearskies.exceptions
from clearskies import authentication, autodoc, typing
from clearskies.endpoints.list import List
from clearskies.functional import string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import Column, Schema, SecurityHeader
    from clearskies.model import Model


class SimpleSearch(List):
    """
    Create an endpoint that supports searching by exact values via url/JSON parameters.

    This acts exactly like the list endpoint but additionally grants the client the ability to search records
    via URL parameters or JSON POST body parameters.  You just have to specify which columns are searchable.

    In the following example we tell the `SimpleSearch` endpoint that we want it to return records from the
    `Student` model, return `id`, `name`, and `grade` in the results, and allow the user to search by
    `name` and `grade`.  We also seed the memory backend with data so the endpoint has something to return:

    ```python
    import clearskies


    class Student(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        grade = clearskies.columns.Integer()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.SimpleSearch(
            Student,
            readable_column_names=["id", "name", "grade"],
            sortable_column_names=["name", "grade"],
            searchable_column_names=["name", "grade"],
            default_sort_column_name="name",
        ),
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": Student,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob", "grade": 5},
                        {"id": "1-2-3-5", "name": "Jane", "grade": 3},
                        {"id": "1-2-3-6", "name": "Greg", "grade": 3},
                        {"id": "1-2-3-7", "name": "Bob", "grade": 2},
                    ],
                },
            ],
        },
    )
    wsgi()
    ```

    Here is the basic operation of the endpoint itself, without any search parameters, in which case it behaves
    identically to the list endpoint:

    ```bash
    $ curl 'http://localhost:8080' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-4",
                "name": "Bob",
                "grade": 5
            },
            {
                "id": "1-2-3-7",
                "name": "Bob",
                "grade": 2
            },
            {
                "id": "1-2-3-6",
                "name": "Greg",
                "grade": 3
            },
            {
                "id": "1-2-3-5",
                "name": "Jane",
                "grade": 3
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```

    We can then search on name via the `name` URL parameter:

    ```bash
    $ curl 'http://localhost:8080?name=Bob' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-4",
                "name": "Bob",
                "grade": 5
            },
            {
                "id": "1-2-3-7",
                "name": "Bob",
                "grade": 2
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```

    and multiple search terms are allowed:

    ```bash
    $ curl 'http://localhost:8080?name=Bob&grade=2' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-7",
                "name": "Bob",
                "grade": 2
            }
        ],
        "pagination": {},
    "input_errors": {}
    }
    ```

    Pagination and sorting work just like with the list endpoint:

    ```bash
    $ curl 'http://localhost:8080?sort=grade&direction=desc&limit=2' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-4",
                "name": "Bob",
                "grade": 5
            },
            {
                "id": "1-2-3-5",
                "name": "Jane",
                "grade": 3
            }
        ],
        "pagination": {
            "number_results": 4,
            "limit": 2,
            "next_page": {
                "start": 2
            }
        },
        "input_errors": {}
    }

    $ curl 'http://localhost:8080?sort=grade&direction=desc&limit=2&start=2' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "1-2-3-6",
                "name": "Greg",
                "grade": 3
            },
            {
                "id": "1-2-3-7",
                "name": "Bob",
                "grade": 2
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```
    """

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        model_class: type[Model],
        readable_column_names: list[str],
        sortable_column_names: list[str],
        searchable_column_names: list[str],
        default_sort_column_name: str,
        default_sort_direction: str = "ASC",
        default_limit: int = 50,
        maximum_limit: int = 200,
        where: typing.condition | list[typing.condition] = [],
        joins: typing.join | list[typing.join] = [],
        url: str = "",
        request_methods: list[str] = ["GET", "POST", "QUERY"],
        response_headers: list[str | Callable[..., list[str]]] = [],
        output_map: Callable[..., dict[str, Any]] | None = None,
        output_schema: Schema | None = None,
        column_overrides: dict[str, Column] = {},
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        authentication: authentication.Authentication = authentication.Public(),
        authorization: authentication.Authorization = authentication.Authorization(),
    ):
        self.request_methods = request_methods

        # we need to call the parent but don't have to pass along any of our kwargs.  They are all optional in our parent, and our parent class
        # just stores them in parameters, which we have already done.  However, the parent does do some extra initialization stuff that we need,
        # which is why we have to call the parent.
        super().__init__(model_class, readable_column_names, sortable_column_names, default_sort_column_name)

    def check_search_in_request_data(self, request_data: dict[str, Any], query_parameters: dict[str, Any]) -> None:
        for input_source_label, input_data in [("request body", request_data), ("URL data", query_parameters)]:
            for column_name, value in input_data.items():
                if column_name in self.allowed_request_keys and column_name not in self.searchable_column_names:
                    continue
                if column_name not in self.searchable_column_names:
                    raise clearskies.exceptions.ClientError(
                        f"Invalid request parameter found in {input_source_label}: '{column_name}'"
                    )
                [relationship_column_name, final_column_name] = self.unpack_column_name_with_relationship(column_name)
                column_to_check = relationship_column_name if relationship_column_name else final_column_name
                value_error = self.searchable_columns[column_to_check].check_search_value(
                    value, relationship_reference=final_column_name
                )
                if value_error:
                    raise clearskies.exceptions.InputErrors({column_name: value_error})

    def configure_model_from_request_data(
        self,
        model: Model,
        request_data: dict[str, Any],
        query_parameters: dict[str, Any],
        pagination_data: dict[str, Any],
    ) -> Model:
        model = super().configure_model_from_request_data(
            model,
            request_data,
            query_parameters,
            pagination_data,
        )

        for input_source in [request_data, query_parameters]:
            for column_name, value in input_source.items():
                if column_name not in self.searchable_column_names:
                    continue

                model = self.add_join(column_name, model)
                [relationship_column_name, column_name] = self.unpack_column_name_with_relationship(column_name)
                if relationship_column_name:
                    self.columns[relationship_column_name].add_search(model, value, relationship_reference=column_name)
                else:
                    model = self.columns[column_name].add_search(model, value, operator="=")

        return model

    def documentation_url_search_parameters(self) -> list[autodoc.request.Parameter]:
        docs = []
        for column in self._get_searchable_columns().values():
            column_doc = column.documentation()
            column_doc.name = self.auto_case_internal_column_name(column_doc.name)
            docs.append(
                autodoc.request.URLParameter(
                    column_doc,
                    description=f"Search by {column_doc.name} (via exact match)",
                )
            )
        return docs  # type: ignore



from typing import Any


class AdditionalConfig:
    """
    This class allows you to add additional names to the Di container.

    The idea here is that you extend the AdditonalConfig class and attach as many
    `provide_*` methods to the class as you want.  This allows the developer to declare a number of
    dependencies and easily attach them to the Di container in one go - helpful for modules that
    come with a variety of things that they want to make available to developers.

    When an AdditionalConfig instance is attached to the Di container, the container (in essence) finds all the
    `provide_*` methods in the class and registers the corresponding name in the Di container - e.g. if you have
    a method called `provide_widget` then when a class or function requests an argument named `widget`
    from the Di container, the container will call the `provide_widget` method on your instance and pass along
    the return value to the thing that requested a `widget`. The `provide_*` methods can declare their own
    dependencies, including ones declared in the same `AdditionalConfig` class (they just can't be circular, of course).

    As always, keep in mind the priority of dependency injection names (see the `clearskies.di.Di` class for full
    details).  If two `AdditionalConfig` objects declare a `provide_*` method with the same name, then the Di
    system will call the method for the `AdditionalConfig` object that was added last.

    By default the Di system caches all values.  If you have a dependency that shouldn't be cached, you can
    extend the `can_cache` method and have it return True/False depending on the name and context.

    Example:
    ```python
    from clearskies.di import Di, AdditionalConfig


    class MyAdditionalConfig(AdditionalConfig):
        def provide_inner_dependency(self):
            return 5

        def provide_important_thing(self, inner_dependency):
            return inner_dependency * 10


    class AnotherAdditionalConfig(AdditionalConfig):
        def provide_inner_dependency(self):
            return 10


    di = Di(additional_configs=[MyAdditionalConfig(), AnotherAdditionalConfig()])
    # Equivalent:
    # di = Di()
    # di.add_addtional_configs([MyAdditionalConfig(), AnotherAdditionalConfig()])


    def my_function(important_thing):
        print(important_thing)  # prints 100
    ```
    """

    def can_cache(self, name: str, di, context: str) -> bool:
        """
        Cache control.

        The Di container caches values by default, but this method allows you to override that.
        After fetching an object from the AdditionalConfig class, the Di container will call this method to
        determine whether or not to cache  it.  `name` will be the name of the Di value that was built, and
        context will be the name of the class that the value was built for.  You then return True/False.

        Note that this controls whether or not to cache the returned value, not whether or not to check the
        cache for a value.  The importance is that, once there is a value in the cache, that will be reused
        for all future requests for that name.  Example:

        ```python
        from clearskies.di import Di, AdditionalConfig
        import secrets


        class MyAdditionalConfig(AdditionalConfig):
            def provide_random_number_not_cached(self):
                return secrets.randbelow(100)

            def provide_random_number_cached(self):
                return secrets.randbelow(100)

            def can_cache(self, name, context=None):
                return name == "random_number_not_cached"


        di = Di(additional_configs=MyAdditionalConfig())


        def my_function(random_number_cached, random_number_not_cached):
            print(random_number_cached)
            print(random_number_not_cached)


        di.call_function(my_function)
        di.call_function(my_function)
        di.call_function(my_function)
        di.call_function(my_function)

        # This prints something like:
        # 58
        # 12
        # 58
        # 14
        # 58
        # 41
        ```
        """
        return True

    def can_build(self, name):
        return hasattr(self, f"provide_{name}")

    def build(self, name, di, context=None):
        if not hasattr(self, f"provide_{name}"):
            raise KeyError(
                f"AdditionalConfig class '{self.__class__.__name__}' cannot build requested dependency, '{name}'"
            )

        return di.call_function(getattr(self, f"provide_{name}"))

    def can_build_class(self, class_to_check: type) -> bool:
        """Return True/False to denote if this AdditionalConfig class can provide a given class."""
        return False

    def build_class(self, class_to_provide: type, argument_name: str, di, context: str = "") -> Any:
        """Return the desired instance of a given class."""
        pass

    def can_cache_class(self, class_to_build: type, di, context: str) -> bool:
        """Control whether or not the Di container caches the instance after building a class."""
        return False



from clearskies.configs import config


class String(config.Config):
    def __init__(self, required=False, default=None, regexp: str = ""):
        self.required = required
        self.default = default
        self.regexp = regexp

    def __set__(self, instance, value: str):
        if not isinstance(value, str):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a parameter that requires a string."
            )
        if self.regexp:
            import re

            if not re.match(self.regexp, value):
                error_prefix = self._error_prefix(instance)
                raise ValueError(
                    f"{error_prefix} attempt to set a value of '{value}' but this does not match the required regexp: '{self.regexp}'."
                )
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> str:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



import json
import sys
from os import isatty
from sys import stdin

from clearskies.input_outputs.input_output import InputOutput


class Cli(InputOutput):
    _args: list[str] = []
    _has_body: bool = False
    _body: str = ""
    _request_method: str = ""
    _request_headers: dict[str, str] = {}

    def __init__(self):
        self._request_headers = {}
        self._args = []
        self._parse_args(sys.argv)
        super().__init__()

    def respond(self, response, status_code=200):
        if status_code != 200:
            sys.exit(response)
        if type(response) != str:
            print(json.dumps(response))
        else:
            print(response)

    def get_arguments(self):
        return sys.argv

    def _parse_args(self, argv):
        tty_data = None
        if not isatty(stdin.fileno()):
            tty_data = sys.stdin.read().strip()

        request_headers = {}
        self._args = []
        kwargs = {}
        index = 0
        # In general we will use positional arguments for routing, and kwargs for request data.
        # If things start with a dash then they are assumed to be a kwarg.  If not, then a positional argument.
        # we don't allow for simple flags: everything is a positional argument or a key/value pair
        # For kwargs, we'll allow for using an equal sign or not, e.g.: '--key=value' or '--key value' or '-d thing'.
        while index < len(argv) - 1:
            index += 1

            # if we don't start with a dash then we are a positional argument
            arg = argv[index]
            if arg[0] != "-":
                self._args.append(arg)
                continue

            # otherwise a kwarg
            arg = arg.strip("-")

            # if we have an equal sign in our kwarg then it's self-contained
            if "=" in arg:
                [key, value] = arg.split("=", 1)

            # otherwise we have to grab the next argument to get the value
            else:
                key = arg
                value = argv[index + 1]
                if "-" in value:
                    raise ValueError(
                        f"Invalid clearskies cli calling sequence: found two key names next to eachother without any values: '-{arg} {value}'"
                    )
                index += 1

            if key.lower() == "h":
                parts = value.split(":", 1)
                if len(parts) != 2:
                    raise ValueError(
                        f"Invalid clearskies cli calling sequence: a parameter named '-H' was found, which is treated as a request header, but it didn't have the proper 'key: value' format."
                    )
                request_headers[parts[0]] = parts[1]
                continue

            kwargs[key] = value

        self._request_headers = request_headers
        self._request_method = "GET"
        request_method_source = ""
        for key in ["x", "X", "request_method"]:
            if key not in kwargs:
                continue

            if request_method_source:
                raise ValueError(
                    f"Invalid clearskies cli calling sequence: the request method was specified via both the -{key} parameter and the -{request_method_source} parameter. To avoid ambiguity, it should only be set once."
                )
            self._request_method = kwargs[key]
            del kwargs[key]
            request_method_source = key

        final_data = None
        data_source = None
        if tty_data:
            final_data = tty_data
            data_source = "piped input"
        if kwargs.get("d"):
            if final_data:
                raise ValueError(
                    f"Invalid clearskies cli calling sequence: request data was sent by both the -d parameter and {data_source}.  To avoid ambiguity, it should only be sent one way."
                )
            final_data = kwargs.get("d")
            data_source = "the -d parameter"
            del kwargs["d"]
        if kwargs.get("data"):
            if final_data:
                raise ValueError(
                    f"Invalid calling sequence: request data was sent by both the -data parameter and {data_source}.  To avoid ambiguity, it should only be sent one way."
                )
            final_data = kwargs.get("data")
            data_source = "the -data parameter"
            del kwargs["data"]
        if final_data and len(kwargs):
            raise ValueError(
                f"Invalid calling sequence: extra parameters were specified after sending a body via {data_source}.  To avoid ambiguity, send all data via {data_source}."
            )
        if not final_data and len(kwargs):
            final_data = kwargs
            data_source = "kwargs"

        # Most of the above inputs result in a string for our final data, in which case we'll leave it as the "raw body"
        # so that it can optionally be interpreted as JSON.  If we received a bunch of kwargs though, we'll allow those to
        # only be "read" as JSON.
        if data_source == "kwargs":
            self._body_as_json = final_data  # type: ignore
            self._body_loaded_as_json = True
            self._has_body = True
            self._body = json.dumps(final_data)
        elif final_data:
            self._has_body = True
            self._body = final_data

    def get_script_name(self):
        return sys.argv[0]

    def get_path_info(self):
        return "/".join(self._args)

    def get_full_path(self):
        return self.get_path_info()

    def get_request_method(self):
        return self._request_method

    def has_body(self):
        return self._has_body

    def get_body(self):
        if not self.has_body():
            return ""

        return self._body

    def context_specifics(self):
        return {}

    def get_client_ip(self):
        return "127.0.0.1"

    def get_query_string(self):
        return ""

    def get_request_headers(self):
        return self._request_headers



import datetime
import inspect
import os
import re
import sys
from types import ModuleType
from typing import Any, Callable

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry  # type: ignore

import clearskies.input_outputs.input_output
import clearskies.secrets
from clearskies.di.additional_config import AdditionalConfig
from clearskies.di.additional_config_auto_import import AdditionalConfigAutoImport
from clearskies.environment import Environment
from clearskies.functional import string


class Di:
    """
    Build a dependency injection object.

    The dependency injection (DI) container is a key part of clearskies, so understanding how to both configure
    them and get dependencies for your classes is important.  Note however that there you don't often have
    to interact with the dependency injection container directly.  All of the configuration options for
    the DI container are also available to all the contexts, which is typically how you will build clearskies
    applications.  So, while you can create a DI container and use it directly, typically you'll just follow
    the same basic techniques to configure your context and use that to run your application.

    These are the main ways to configure the DI container:

     1. Import classes - each imported class is assigned an injection name based on the class name.
     2. Import modules - clearskies will iterate over the module and import all the classes and AdditionalConfigAutoImport classes it finds.
     3. Import AdditionalConfig classes - these allow you to programmatically define dependencies.
     4. Specify bindings - this allows you to provide any kind of value with whatever name you want.
     5. Specify class overrides - these allow you to swap out classes directly.
     6. Extending the Di class - this allows you to provide a default set of values.

    When the DI system builds a class or calls a function, those classes and functions can themselves request any value
    configured inside the DI container.  There are three ways to request the desired dependencies:

     1. By type hinting a class on any arguments (excluding python built-ins)
     2. By specifying the name of a registered dependency
     3. By extending the `clearskies.di.AutoFillProps` class and creating class properties from the `clearskies.di.inject_from` module

    Note that when a class is built/function is called by the DI container, keyword arguments are not allowed
    (because the DI container doesn't know whether or not it should provide optional arguments).  In addition,
    the DI container must be able to resolve all positional arguments.  If the class requests an argument
    that the DI system does not recognize, an error will be thrown.  Finally, it's a common pattern in clearskies
    for some portion of the system to accept functions that will be called by the DI container.  When this happens,
    it's possible for clearskies to provide additional values that may be useful when executing the function.
    The areas that accept functions like this also document the additional dependency injection names that are available.

    Given the variety of ways that dependencies can be specified, it's important to understand the order the priority that
    clearskies uses to determine what value to provide in case there is more than one source.  That order is:

     1. Positional arguments with type hints:
        1. The override class if the type-hinted class has a registered override
        2. A value provided by an AdditionalConfig that can provide the type-hinted class
        3. The class itself if the class has been added explicitly via add_classes or implicitly via add_modules
        4. A clearskies built-in for predefined types
     2. All other positional arguments will have values provided based on the argument name and will receive
        1. Things set via `add_binding(name, value)`
        2. Class added via `add_classes` or `add_modules` which are made available according to their Di name
        3. An AdditionalConfig class with a corresponding `provide_[name]` function
        4. A clearskies built-in for predefined names

    Here is the list of predefined values with their names and types:

    | Injection Name       | Injection Type                                          | Value                                                                                     |
    |----------------------|---------------------------------------------------------|-------------------------------------------------------------------------------------------|
    | di                   | -                                                       | The active Di container                                                                   |
    | now                  | -                                                       | The current time in a datetime object, without timezone                                   |
    | utcnow               | -                                                       | The current time in a datetime object, with timezone set to UTC                           |
    | requests             | requests.Session                                        | A requests object configured to allow a small number of retries                           |
    | input_output         | clearskies.input_outputs.InputOutput                    | The clearskies builtin used for receiving and sending data to the client                  |
    | uuid                 | -                                                       | `import uuid` - the uuid module builtin to python                                         |
    | environment          | clearskies.Environment                                  | A clearskies helper that access config info from the environment or a .env file           |
    | sys                  | -                                                       | `import sys` - the sys module builtin to python                                           |
    | oai3_schema_resolver | -                                                       | Used by the autodoc system                                                                |
    | connection_details   | -                                                       | A dictionary containing credentials that pymysql should use when connecting to a database |
    | connection           | -                                                       | A pymysql connection object                                                               |
    | cursor               | -                                                       | A pymysql cursor object                                                                   |

    Note: for dependencies with an injection name but no injection type, this means that to inject those values you
    must name your argument with the given injection name.  In all of the above cases though you can still add type
    hints if desired.  So, for instance, you can declare an argument of `utcnow: datetime.datetime`.  clearskies
    will ignore the type hint (since `datetime.datetime` isn't a type with a predefined value in clearskies) and
    identify the value based on the name of the argument.

    Note: multiple `AdditionalConfig` classes can be added to the Di container, and so a single injection name or class
    can potentially be provided by multiple AdditionalConfig classes.  AdditionalConfig classes are checked in the
    reverse of the order they were addded in - classes added last are checked first when trying to find values.

    Note: When importing modules, any classes that inherit from `AdditionalConfigAutoImport` are automatically added
    to the list of additional config classes.  These classes are added at the top of the list, so they are lower
    priority than any classes you add via `add_additional_configs` or the `additional_configs` argument of the Di
    constructor.

    Note: Once a value is constructed, it is cached by the Di container and will automatically be provided for future
    references of that same Di name or class.  Arguments injected in a constructor will always receive the cached
    value.  If you want a "fresh" value of a given dependency, you have to attach instances from the
    `clearskies.di.inject` module onto class proprties.  The instances in the `inject` module generally
    give options for cache control.

    Here's an example that brings most of these pieces together.  Once again, note that we're directly using
    the Di contianer to build class/call functions, while normally you configure the Di container via your context
    and then clearskies itself will build your class or call your functions as needed.  Full explanation comes after
    the example.

    ```python
    from clearskies.di import Di, AdditionalConfig


    class SomeClass:
        def __init__(self, my_value: int):
            self.my_value = my_value


    class MyClass:
        def __init__(self, some_specific_value: int, some_class: SomeClass):
            # `some_specific_value` is defined in both `MyProvider` and `MyOtherProvider`
            # `some_class` will be injected from the type hint, and the actual instance is made by our
            # `MyProvider`
            self.final_value = some_specific_value * some_class.my_value


    class VeryNeedy:
        def __init__(self, my_class, some_other_value: str):
            # We're relying on the automatic conversion of class names to snake_case, so clearskies
            # will connect `my_class` to `MyClass`, which we provided directly to the Di container.

            # some_other_value is specified as a binding
            self.my_class = my_class
            self.some_other_value = some_other_value


    class MyOtherProvider(AdditionalConfig):
        def provide_some_specific_value(self):
            # the order of additional configs will cause this function to be invoked
            # (and hence some_specific_value will be `10`) despite the fact that MyProvider
            # also has a `provide_` function with the same name.
            return 10


    class MyProvider(AdditionalConfig):
        def provide_some_specific_value(self):
            # note that the name of our function matches the name of the argument
            # expected by MyClass.__init__.  Again though, we won't get called because
            # the order the AdditionalConfigs are loaded gives `MyOtherProvider` priority.
            return 5

        def can_build_class(self, class_to_check: type) -> bool:
            # this lets the DI container know that if someone wants an instance
            # of SomeClass, we can build it.
            return class_to_check == SomeClass

        def build_class(self, class_to_provide: type, argument_name: str, di, context: str = ""):
            if class_to_provide == SomeClass:
                return SomeClass(5)
            raise ValueError(
                f"I was asked to build a class I didn't expect '{class_to_provide.__name__}'"
            )


    di = Di(
        classes=[MyClass, VeryNeedy, SomeClass],
        additional_configs=[MyProvider(), MyOtherProvider()],
        bindings={
            "some_other_value": "dog",
        },
    )


    def my_function(my_fancy_argument: VeryNeedy):
        print(f"Jane owns {my_fancy_argument.my_class.final_value}:")
        print(f"{my_fancy_argument.some_other_value}s")


    print(di.call_function(my_function))
    # prints 'Jane owns 50 dogs'
    ```

    When `call_function` is executed on `my_function`, the di system checks the calling arguments of `my_function`
    and runs through the priority list above to populate them.  `my_function` has one argument -
    `my_fancy_argument: VeryNeedy`, which it resolves as so:

     1. The type hint (`VeryNeedy`) matches an imported class.  Therefore, clearskies will build an instance of VeryNeedy and
        provide it for `my_fancy_argument`.
     2. clearskies inpsects the constructor for `VeryNeedy` and finds two arguments, `my_class` and `some_other_value: str`,
        which it attempts to build.
        1. `my_class` has no type hint, so clearskies falls back on name-based resolution.  A class called `MyClass` was imported,
           and per standard naming convention, this automatically becomes available via the name `my_class`.  Thus, clearskies
           prepares to build an instance of `MyClass`. `MyClass` has two arguments: `some_specific_value: int` and
           `some_class: SomeClass`
           1. For `some_specific_value`, the Di service falls back on named-based resolution (because it will never try to
              provide values for type-hints of built-in types).  Both `MyOtherProvider` and `MyProvider` have a method called
              `provide_some_specific_value`, so both can be used to provide this value.  Since `MyOtherProvider` was added to
              the Di container last, it takes priority.  Therefore, clearskies calls `MyOtherProvider.provide_some_specific_value`
              to create the value that it will populate into the `some_specific_value` parameter.
           2. For `some_class: SomeClass`, clearskies evaluates the type-hint.  It works through the additional configs and, since
              `MyProvider` returns True when `can_build_class` is called with  `SomeClass`, the Di container will use this
              additional config to create the value for the `some_class` argument.  Therefore, clearskies calls
              `MyProvider.build_class(SomeClass, 'some_class', di)` and the return value is used for the `some_class` argument.
        2. `some_other_value` uses a built-in for a type hint, so clearskies falls back on name-based resolution.  It falls back
           on the registered binding of `"dog"` to the name `"some_other_value"`, so clearskies provides `"dog"`.
    """

    _added_modules: dict[int, bool] = {}
    _additional_configs: list[AdditionalConfig] = []
    _bindings: dict[str, Any] = {}
    _building: dict[int, str] = {}
    _classes: dict[str, dict[str, int | type]] = {}
    _prepared: dict[str, Any] = {}
    _class_overrides_by_name: dict[str, type] = {}
    _class_overrides_by_class: dict[type, Any] = {}
    _type_hint_disallow_list = [int, float, str, dict, list, datetime.datetime]
    _now: datetime.datetime | None = None
    _utcnow: datetime.datetime | None = None
    _predefined_classes_name_map: dict[type, str] = {
        requests.Session: "requests",
        clearskies.input_outputs.input_output.InputOutput: "input_output",
        Environment: "environment",
    }

    def __init__(
        self,
        classes: type | list[type] = [],
        modules: ModuleType | list[ModuleType] = [],
        bindings: dict[str, Any] = {},
        additional_configs: AdditionalConfig | list[AdditionalConfig] = [],
        class_overrides: dict[type, Any] = {},
        overrides: dict[str, type] = {},
        now: datetime.datetime | None = None,
        utcnow: datetime.datetime | None = None,
    ):
        """
        Create a dependency injection container.

        For details on the parameters, see the related methods:

        classes -> di.add_classes()
        modules -> di.add_modules()
        bindings -> di.add_binding()
        additional_configs -> di.add_additional_configs()
        class_overrides -> di.add_class_override()
        """
        self._added_modules = {}
        self._additional_configs = []
        self._bindings = {}
        self._building = {}
        self._classes = {}
        self._class_overrides_by_name = {}
        self._class_overrides_by_class = {}
        self._prepared = {}
        if classes is not None:
            self.add_classes(classes)
        if modules is not None:
            self.add_modules(modules)
        if bindings is not None:
            for key, value in bindings.items():
                self.add_binding(key, value)
        if additional_configs is not None:
            self.add_additional_configs(additional_configs)
        if class_overrides:
            for key, value in class_overrides.items():  # type: ignore
                self.add_class_override(key, value)  # type: ignore
        if overrides:
            for key, value in overrides.items():
                self.add_override(key, value)
        if now:
            self.set_now(now)
        if utcnow:
            self.set_utcnow(utcnow)

    def add_classes(self, classes: type | list[type]) -> None:
        """
        Record any class that should be made available for injection.

        All classes that come in here become available via their injection name, which is calculated
        by converting the class name from TitleCase to snake_case.  e.g. the following class:

        ```python
        class MyClass:
            pass
        ```

        gets an injection name of `my_class`.  Also, clearskies will only resolve and reject based on type hints
        if those classes are first added via `add_classes`.  See the following example:

        ```python
        from clearskies.di import Di

        class MyClass:
            name = "Simple Demo"

        di = Di(classes=[MyClass])
        # equivalent: di.add_classes(MyClass), di.add_classes([MyClass])

        def my_function(my_class):
            print(my_class.name)

        def my_function_with_type_hinting(the_name_no_longer_matters: MyClass):
            print(my-class.name)

        # both print 'Simple Demo'
        di.call_function(my_function)
        di.call_function(my_function_with_type_hinting)
        ```
        """
        if not isinstance(classes, list):
            classes = [classes]
        for add_class in classes:
            name = string.camel_case_to_snake_case(add_class.__name__)
            self._classes[name] = {"id": id(add_class), "class": add_class}
            self._classes[add_class] = {"id": id(add_class), "class": add_class}  # type: ignore

            # if this is a model class then also add a plural version of its name
            # to the DI configuration
            if hasattr(add_class, "id_column_name"):
                self._classes[string.make_plural(name)] = {"id": id(add_class), "class": add_class}

    def add_modules(
        self, modules: ModuleType | list[ModuleType], root: str | None = None, is_root: bool = True
    ) -> None:
        """
        Add a module to the dependency injection container.

        clearskies will iterate through the module, adding all imported classes into the dependency injection container.

        So, consider the following file structure inside a module:

        ```
        my_module/
            __init__.py
            my_sub_module/
                __init__.py
                my_class.py
        ```

        Assuming that the submodule and class are imported at each level (e.g. my_module/__init__.py imports my_sub_module,
        and my_sub_module/__init__.py imports my_class.py) then you can:

        ```python
        from clearksies.di import Di
        import my_module

        di = Di()
        di.add_modules([
            my_module
        ])  # also equivalent: di.add_modules(my_module), or Di(modules=[my_module])


        def my_function(my_class):
            pass


        di.call_function(my_function)
        ```

        `my_function` will be called and `my_class` will automatically be populated with an instance of
        `my_module.sub_module.my_class.MyClass`.

        Note that MyClass will be able to declare its own dependencies per normal dependency injection rules.
        See the main docblock in the clearskies.di.Di class for more details about how all the pieces work together.
        """
        if not isinstance(modules, list):
            modules = [modules]

        for module in modules:
            # skip internal python modules
            if not hasattr(module, "__file__") or not module.__file__:
                continue
            module_id = id(module)
            if is_root:
                root = os.path.dirname(module.__file__)
            root_len = len(root) if root else 0
            if module_id in self._added_modules:
                continue
            self._added_modules[module_id] = True

            for name, item in module.__dict__.items():
                if inspect.isclass(item):
                    try:
                        class_root = os.path.dirname(inspect.getfile(item))
                    except TypeError:
                        # built-ins will end up here
                        continue
                    if class_root[:root_len] != root:
                        continue
                    if issubclass(item, AdditionalConfigAutoImport):
                        init_args = inspect.getfullargspec(item)
                        nargs = len(init_args.args) if init_args.args else 0
                        nkwargs = len(init_args.defaults) if init_args.defaults else 0
                        if nargs - 1 - nkwargs > 0:
                            raise ValueError(
                                "Error auto-importing additional config "
                                + item.__name__
                                + ": auto imported configs can only have keyword arguments."
                            )
                        self.add_additional_configs([item()])
                        continue
                    self.add_classes([item])
                if inspect.ismodule(item):
                    if not hasattr(item, "__file__") or not item.__file__:
                        continue
                    child_root = os.path.dirname(item.__file__)
                    if child_root[:root_len] != root:
                        continue
                    if module.__name__ == "clearskies":
                        break
                    self.add_modules([item], root=root, is_root=False)

    def add_additional_configs(self, additional_configs: AdditionalConfig | list[AdditionalConfig]) -> None:
        """
        Add an additional config instance to the dependency injection container.

        Additional config class provide an additional way to provide dependencies into the dependency
        injection system.  For more details about how to use them, see both base classes:

         1. clearskies.di.additional_config.AdditionalConfig
         2. clearskies.di.additional_config_auto_import.AdditionalConfigAutoImport

        To use this method:

        ```python
        import clearskies.di


        class MyConfig(clearskies.di.AdditionalConfig):
            def provide_some_value(self):
                return 2

            def provide_another_value(self, some_value):
                return some_value * 2


        di = clearskies.di.Di()
        di.add_additional_configs([MyConfig()])
        # equivalents:
        # di.add_additional_configs(MyConfig())
        # di = clearskies.di.Di(additional_configs=[MyConfig()])


        def my_function(another_value):
            print(another_value)  # prints 4


        di.call_function(my_function)
        ```
        """
        if not isinstance(additional_configs, list):
            additional_configs = [additional_configs]
        self._additional_configs.extend(additional_configs)

    def add_binding(self, key, value):
        """
        Provide a specific value for name-based injection.

        This method attaches a value to a specific dependency injection name.

        ```python
        import clearskies.di

        di = clearskies.di.Di()
        di.add_binding("my_name", 12345)
        # equivalent:
        # di = clearskies.di.Di(bindings={"my_name": 12345})


        def my_function(my_name):
            print(my_name)  # prints 12345


        di.call_function(my_function)
        ```
        """
        if key in self._building:
            raise KeyError(f"Attempt to set binding for '{key}' while '{key}' was already being built")

        # classes are placed in self._bindings, but any other prepared value goes straight into self._prepared
        if inspect.isclass(value):
            self._bindings[key] = value
            if key in self._prepared:
                del self._prepared[key]
        else:
            self._prepared[key] = value

    def add_class_override(self, class_to_override: type, replacement: Any) -> None:
        """
        Override a class for type-based injection.

        This function allows you to replace/mock class provided when relying on type hinting for injection.
        This is most often (but not exclusively) used for mocking out classes during texting.  Note that
        this only overrides that specific class - not classes that extend it.

        Example:
        ```python
        from clearskies.import Di

        class TypeHintedClass:
            my_value = 5

        class ReplacementClass:
            my_value = 10

        di = Di()
        di.add_classes(TypeHintedClass)
        di.add_class_override(TypeHintedClass, ReplacementClass)
        # also di = Di(class_overrides={TypeHintedClass: ReplacementClass})

        def my_function(some_value: TypeHintedClass):
            print(some_value.my_value) # prints 10

        di.call_function(my_function)
        ```
        """
        if not inspect.isclass(class_to_override):
            raise ValueError(
                "Invalid value passed to add_class_override for 'class_or_name' parameter: it was neither a name nor a class"
            )

        self._class_overrides_by_class[class_to_override] = replacement

    def has_class_override(self, class_to_check: type) -> bool:
        return class_to_check in self._class_overrides_by_class

    def get_override_by_class(self, object_to_override: Any) -> Any:
        if object_to_override.__class__ not in self._class_overrides_by_class:
            return object_to_override

        override = self._class_overrides_by_class[object_to_override.__class__]
        if inspect.isclass(override):
            return self.build_class(override)
        if hasattr(override, "injectable_properties"):
            override.injectable_properties(self)
        return override

    def add_override(self, name: str, replacement_class: type) -> None:
        """Override a specific injection name by specifying a class that should be injected in its place."""
        if not inspect.isclass(replacement_class):
            raise ValueError(
                "Invalid value passed to add_override for 'replacement_class' parameter: a class should be passed but I got a "
                + str(type(replacement_class))
            )

        self._class_overrides_by_name[name] = replacement_class

    def set_now(self, now: datetime.datetime) -> None:
        """Set the current time which will be passed along to any dependency arguments named `now`."""
        if now.tzinfo is not None:
            raise ValueError(
                "set_now() was passed a datetime object with timezone information - it should only be given timezone-naive datetime objects.  Maybe you meant to use di.set_utcnow()"
            )
        self._now = now

    def set_utcnow(self, utcnow: datetime.datetime) -> None:
        """Set the current time which will be passed along to any dependency arguments named `utcnow`."""
        if not utcnow.tzinfo:
            raise ValueError(
                "set_utcnow() was passed a datetime object without timezone information - it should only be given timezone-aware datetime objects.  Maybe you meant to use di.set_now()"
            )
        self._utcnow = utcnow

    def build(self, thing: Any, context: str | None = None, cache: bool = False) -> Any:
        """
        Have the dependency injection container build a value for you.

        This will accept either a dependency injection name or a class.
        """
        if inspect.isclass(thing):
            return self.build_class(thing, context=context, cache=cache)
        elif type(thing) == str:
            return self.build_from_name(thing, context=context, cache=cache)
        elif callable(thing):
            raise ValueError("build received a callable: you probably want to use di.call_function()")

        # if we got here then our thing is already and object of some sort and doesn't need anything further
        return thing

    def build_from_name(self, name: str, context: str | None = None, cache: bool = False) -> Any:
        """
        Build a dependency based on its name.

        Order of priority:
          1. Things set via `add_binding(name, value)`
          2. Class added via `add_classes` or `add_modules` which are made available according to their Di name
          3. An AdditionalConfig class with a corresponding `provide_[name]` function
          4. The Di class itself if it has a matching `provide_[name]` function (aka the builtins)
        """
        if name in self._prepared and cache:
            return self._prepared[name]

        if name in self._bindings:
            built_value = self.build(self._bindings[name], context=context)
            if cache:
                self._prepared[name] = built_value
            return built_value

        if name in self._classes or name in self._class_overrides_by_name:
            class_to_build = (
                self._class_overrides_by_name[name]
                if name in self._class_overrides_by_name
                else self._classes[name]["class"]
            )  # type: ignore
            built_value = self.build_class(class_to_build, context=context)  # type: ignore
            if cache:
                self._prepared[name] = built_value  # type: ignore
            return built_value

        # additional configs are meant to override ones that come before, with most recent ones
        # taking precedence.  Therefore, start at the end (e.g. FILO instead of FIFO, except nothing actually leaves)
        for index in range(len(self._additional_configs) - 1, -1, -1):
            additional_config = self._additional_configs[index]
            if not additional_config.can_build(name):
                continue
            built_value = additional_config.build(name, self, context if context else "")
            if cache and additional_config.can_cache(name, self, context if context else ""):
                self._prepared[name] = built_value
            return built_value

        if hasattr(self, f"provide_{name}"):
            built_value = self.call_function(getattr(self, f"provide_{name}"))
            if cache and self.can_cache(name, context if context else ""):
                self._prepared[name] = built_value
            return built_value

        # why twice?  When a "concrete" value is bound directly to a DI name, it is just
        # put directly in the cache.  Therefore, if cache=False, we won't find it (which is a bug).
        # Therefore, if we get to the very bottom, haven't found anything, but it is in the
        # cache: well, it's time to use the cache.
        if name in self._prepared:
            return self._prepared[name]

        context_note = f" for {context}" if context else ""
        raise ValueError(
            f"I was asked to build {name}{context_note} but there is no added class, configured binding, "
            + f"or a corresponding 'provide_{name}' method for this name."
        )

    def build_argument(self, argument_name: str, type_hint: type | None, context: str = "", cache: bool = True) -> Any:
        """
        Build an argument given the name and type hint.

        Runs through the resolution order described in the docblock at the top of the Di class to build an argument given
        its name and type-hint.
        """
        built_value = self.build_class_from_type_hint(argument_name, type_hint, context=context, cache=True)
        if built_value:
            return built_value
        return self.build_from_name(argument_name, context=context, cache=True)

    def build_class(self, class_to_build: type, context=None, cache=True) -> Any:
        """
        Build a class.

        The class constructor cannot accept any kwargs.   See self._disallow_kwargs for more details
        """
        if class_to_build in self._prepared and cache:
            return self._prepared[class_to_build]  # type: ignore
        my_class_name = class_to_build.__name__

        init_args = inspect.getfullargspec(class_to_build)
        if init_args.defaults is not None:
            self._disallow_kwargs(f"build class '{class_to_build.__name__}'")

        # ignore the first argument because that is just `self`
        build_arguments = init_args.args[1:]
        if not build_arguments:
            if hasattr(class_to_build, "injectable_properties"):
                class_to_build.injectable_properties(self)
            built_value = class_to_build()
            if cache:
                self._prepared[class_to_build] = built_value  # type: ignore
            return built_value

        # self._building will help us keep track of what we're already building, and what we are building it for.
        # This is partly to give better error messages, but mainly to detect cyclical dependency trees.
        # We use id(class_to_build) because this uniquely identifies a class (while the name won't, since two
        # different classes can have the same name but be in different modules).  Therefore, before we start
        # building a class, see if its id is in self._building, raise an error if so, or continue if not.
        class_id = id(class_to_build)
        if class_id in self._building:
            original_context_label = (
                f"'{self._building[class_id]}'" if self._building[class_id] is not None else "itself"
            )
            raise ValueError(
                f"Circular dependencies detected while building '{my_class_name}' because '"
                + f"{my_class_name} is a dependency of both '{context}' and {original_context_label}"
            )

        self._building[class_id] = context
        # Turn on caching when building the automatic dependencies that get injected into a class constructor
        args = [
            self.build_argument(
                build_argument, init_args.annotations.get(build_argument, None), context=my_class_name, cache=True
            )
            for build_argument in build_arguments
        ]

        del self._building[class_id]

        built_value = class_to_build(*args)
        if hasattr(built_value, "injectable_properties"):
            built_value.injectable_properties(self)
        if cache:
            self._prepared[class_to_build] = built_value  # type: ignore
        return built_value

    def build_class_from_type_hint(
        self, argument_name: str, class_to_build: type | None, context: str = "", cache: bool = True
    ) -> Any | None:
        """
        Build an argument from a type hint.

        Note that in many cases we can't actually build the thing.  It may be a type hint of a built-in or some other value
        that we're not configured to deal with.  In that case, just return None and the calling method will deal with it.

        This follows the resolution order defined in the docblock of the Di class.
        """
        # these first checks just verify that it is something that we can actually build
        if not class_to_build:
            return None
        if not callable(class_to_build):
            return None
        if inspect.isabstract(class_to_build):
            return None

        # then first things first: check our class overrides
        if class_to_build in self._class_overrides_by_class:
            return self.build_class(self._class_overrides_by_class[class_to_build], context=context, cache=cache)

        # next check our additional config classes
        built_value = None
        can_cache = False
        for index in range(len(self._additional_configs) - 1, -1, -1):
            additional_config = self._additional_configs[index]
            if not additional_config.can_build_class(class_to_build):
                continue

            built_value = additional_config.build_class(class_to_build, argument_name, self, context=context)
            can_cache = additional_config.can_cache_class(class_to_build, self, context)
            break

        # a small handful of predefined classes
        if class_to_build in self._predefined_classes_name_map:
            dependency_name = self._predefined_classes_name_map[class_to_build]
            built_value = self.call_function(getattr(self, f"provide_{dependency_name}"))
            can_cache = self.can_cache(dependency_name, context if context else "")

        # finally, if we found something, cache and/or return it
        if built_value is not None:
            if cache and can_cache:
                self._prepared[class_to_build] = built_value  # type: ignore
            return built_value

        # last but not least we build the class itself as long as it has been imported into the Di system
        if class_to_build in self._classes:
            return self.build_class(class_to_build, context=context, cache=cache)

        return None

    def call_function(self, callable_to_execute: Callable, **kwargs):
        """
        Call a function, building any positional arguments and providing them.

        Any kwargs passed to call_function will populate the equivalent dependencies.

        ```python
        from clearskies.di import Di

        di = Di(bindings={"some_name": "hello"})


        def my_function(some_name, some_other_name):
            print(f"{some_name} {some_other_value}")  # prints 'hello world'


        di.call_function(my_function, some_other_value="world")
        ```
        """
        args_data = inspect.getfullargspec(callable_to_execute)

        # we need to decide if we've been passed a bound method, because then we need to ignore the
        # first argument (aka `self`).  The simplest way to do this is to check for the `__self__` attr,
        # but this will be fooled by methods with decorators.  There doesn't seem to be a good solution to this
        # that works in all cases: https://stackoverflow.com/a/50074581/1921979
        call_arguments = args_data.args
        if hasattr(callable_to_execute, "__self__"):
            call_arguments = call_arguments[1:]

        # separate out args and kwargs.  kwargs for the function are only allowed to come out of the kwargs
        # we were passed.  If the function has a kwarg that we don't have, then ignore it.
        # args come out of dependencies or the kwargs passed to us.  If an arg is missing, then throw an error.
        nargs = len(call_arguments)
        nkwargs = len(args_data.defaults) if args_data.defaults else 0
        arg_names = call_arguments[: nargs - nkwargs]
        kwarg_names = call_arguments[nargs - nkwargs :]

        callable_args = [
            (
                kwargs[arg]
                if arg in kwargs
                else self.build_argument(
                    arg, args_data.annotations.get(arg, None), context=callable_to_execute.__name__, cache=True
                )
            )
            for arg in arg_names
        ]
        callable_kwargs = {}
        for kwarg_name in kwarg_names:
            if kwarg_name not in kwargs:
                continue
            callable_kwargs[kwarg_name] = kwargs[kwarg_name]

        return callable_to_execute(*callable_args, **callable_kwargs)

    def _disallow_kwargs(self, action):
        """
        Raise an exception.

        This is used to raise an exception and stop building a class if its constructor accepts kwargs. To be clear,
        we actually can support kwargs - it just doesn't make much sense.  The issue is that keywords are
        optional, so a dependency injection framework doesn't know what to do with them.  Ignore them?  Provide them?
        The behavior is unclear, and therefore, bug prone.

        If you need to create a class that accepts kwargs in its constructor you can do  that though - just override
        this method in your DI class and don't raise an exception.  If you do that, everything will still work but
        nothing will be provided for your kwargs.

        Another option would be to do a self.build_from_name on the kwargs names, but only provide the kwarg
        if the name is something that the DI container can actually provide (and otherwise just let it fall back
        on the default).  However, I'm not convinced that will be a helpful use case, so I'm not implementing
        it right now.
        """
        raise ValueError(f"Cannot {action} because it has keyword arguments.")

    def can_cache(self, name: str, context: str) -> bool:
        """Control whether or not to cache a value built by the DI container."""
        if name == "now" or name == "utcnow":
            return False
        return True

    def provide_di(self):
        return self

    def provide_requests(self):
        retry_strategy = Retry(
            total=3,
            status_forcelist=[429, 500, 502, 503, 504],
            backoff_factor=1,
            allowed_methods=["GET", "POST", "DELETE", "OPTIONS", "PATCH"],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session = requests.Session()
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        return session

    def provide_sys(self):
        import sys

        return sys

    def provide_environment(self):
        return Environment(os.getcwd() + "/.env", os.environ, {})

    def provide_connection_no_autocommit(self, connection_details):
        # I should probably just switch things so that autocommit is *off* by default
        # and only have one of these, but for now I'm being lazy.
        try:
            import pymysql
        except:
            raise ValueError(
                "The cursor requires pymysql to be installed.  This is an optional dependency of clearskies, so to include it do a `pip install 'clear-skies[mysql]'`"
            )

        return pymysql.connect(
            user=connection_details["username"],
            password=connection_details["password"],
            host=connection_details["host"],
            database=connection_details["database"],
            port=connection_details.get("port", 3306),
            ssl_ca=connection_details.get("ssl_ca", None),
            autocommit=False,
            connect_timeout=2,
            cursorclass=pymysql.cursors.DictCursor,
        )

    def provide_connection(self, connection_details):
        try:
            import pymysql
        except:
            raise ValueError(
                "The cursor requires pymysql to be installed.  This is an optional dependency of clearskies, so to include it do a `pip install 'clear-skies[mysql]'`"
            )

        return pymysql.connect(
            user=connection_details["username"],
            password=connection_details["password"],
            host=connection_details["host"],
            database=connection_details["database"],
            port=connection_details.get("port", 3306),
            ssl_ca=connection_details.get("ssl_ca", None),
            autocommit=True,
            connect_timeout=2,
            cursorclass=pymysql.cursors.DictCursor,
        )

    def provide_connection_details(self, environment):
        try:
            port = environment.get("db_port")
        except:
            port = 3306

        try:
            ssl_ca = environment.get("db_ssl_ca")
        except:
            ssl_ca = None

        return {
            "username": environment.get("db_username"),
            "password": environment.get("db_password"),
            "host": environment.get("db_host"),
            "database": environment.get("db_database"),
            "port": port,
            "ssl_ca": ssl_ca,
        }

    def provide_cursor(self, connection):
        return connection.cursor()

    def provide_now(self):
        return datetime.datetime.now() if self._now is None else self._now

    def provide_utcnow(self):
        return datetime.datetime.now(datetime.timezone.utc) if self._utcnow is None else self._utcnow

    def provide_input_output(self):
        raise AttributeError(
            "The dependency injector requested an InputOutput but none has been configured.  Alternatively, if you directly called `di.build('input_output')` then try again with `di.build('input_output', cache=True)`"
        )

    def provide_oai3_schema_resolver(self):
        from clearskies import autodoc

        return autodoc.formats.oai3_json.OAI3SchemaResolver()

    def provide_uuid(self):
        import uuid

        return uuid

    def provide_secrets(self):
        return clearskies.secrets.Secrets()

    def provide_memory_backend_default_data(self):
        return []

    def provide_global_table_prefix(self):
        return ""

    def provide_akeyless(self):
        import akeyless  # type: ignore

        return akeyless



"""
The backend system provides a layer of abstraction between a clearskies model and the place where it stores its data.

## Motivation

The goal of the backend system is to create a standard way to interact with external systems, whether those are
databases, APIs, cloud resources, etc...  For comparison, models in other frameworks typically interact exclusively
with SQL databases.  Therefore, building integrations with other data stores or 3rd party systems requires the
creation of bespoke SDKs, which results in time spent integrating and learning how to use them.  Clearskies models
work with different backends to normalize the process of integrating with data stores and external systems.
This ability to switch backends turns the models, in essence, into a standardized SDK.  The details of how to
interact with some other system are abstracted away into the backend, and then developers can interact with the
system using models in the standard way.

Of course, no abstraction layer is perfect, and not all datastores can fit into the backend system provided by
clearskies.  That's okay!  Like every abstraction layer in clearskies, you should use it when it works for
you and forget about it otherwise.  So, for instance, it is not possible to build arbitrarily complex SQL
queries with the query system provided by models, but you can always inject the cursor and use it directly.
If some arbitrary data store just works in a completely different way, such that the backend system doesn't
work for it, it's still fine to build an SDK the "old school" way and use it in a clearskies application.

## Backends

There are four primary kinds of backends built into clearskies, with a few sub-types.  Those are:

 1. CursorBackend: Store data in an SQL-like database
 2. MemoryBackend: Store data in-memory
 3. ApiBackend: Use an API to fetch and store data
    1. ApiGetOnlyBackend: Only supports fetching individual records from the API by id.
    2. RestfulApiAdvancedSearchBackend: Automatically integrates with APIs built using clearskies.endpoints.RestfulApi
 4. SecretsBackend - Store data directly in a secrets manager

See the documentation for each backend to understand how to configure and use it.  In all cases though, you specify
the backend by instantiating the backend and attaching it to the model class via the `backend` attribute:

```python
import clearskies


class MyModel(clearskies.model):
    backend = clearskies.backends.MemoryBackend()
```
"""

# from .api_backend import ApiBackend
# from .api_get_only_backend import ApiGetOnlyBackend
from clearskies.backends.api_backend import ApiBackend
from clearskies.backends.backend import Backend
from clearskies.backends.cursor_backend import CursorBackend
from clearskies.backends.memory_backend import MemoryBackend

# from .memory_backend import MemoryBackend
# from .restful_api_advanced_search_backend import RestfulApiAdvancedSearchBackend
# from .secrets_backend import SecretsBackend


__all__ = [
    "ApiBackend",
    # "ApiGetOnlyBackend",
    "Backend",
    "CursorBackend",
    "MemoryBackend",
    # "RestfulApiAdvancedSearchBackend",
    # "SecretsBackend",
]



from typing import Any

from clearskies.configs import config


class Configurable:
    _config: dict[str, Any] | None = None
    _descriptor_config_map: dict[int, str] | None = None

    def _set_config(self, descriptor, value):
        if not self._config:
            self._config = {}

        self._config[self._descriptor_to_name(descriptor)] = value

    def _get_config(self, descriptor):
        if not self._config:
            self._config = {}

        name = self._descriptor_to_name(descriptor)
        if name not in self._config:
            raise KeyError(f"Attempt to fetch a config value named '{name}' but no value has been set for this config")
        return self._config[name]

    @classmethod
    def _get_config_object(cls, attribute_name):
        return getattr(cls, attribute_name)

    @classmethod
    def get_descriptor_config_map(cls):
        if cls._descriptor_config_map:
            return cls._descriptor_config_map

        descriptor_config_map = {}
        for attribute_name in dir(cls):
            descriptor = getattr(cls, attribute_name)
            if not isinstance(descriptor, config.Config):
                continue

            descriptor_config_map[id(descriptor)] = attribute_name

        cls._descriptor_config_map = descriptor_config_map
        return cls._descriptor_config_map

    def _descriptor_to_name(self, descriptor):
        descriptor_config_map = self.get_descriptor_config_map()
        if id(descriptor) not in descriptor_config_map:
            raise ValueError(
                f"The reason behind this error is kinda long and complicated, but doens't really matter.  To make it go away, just add `_descriptor_config_map = None` to the definition of {self.__class__.__name__}"
            )
        return descriptor_config_map[id(descriptor)]

    def finalize_and_validate_configuration(self):
        my_class = self.__class__
        if not self._config:
            self._config = {}

        # now it's time to check for required values and provide defaults
        attribute_names = self.get_descriptor_config_map().values()
        for attribute_name in attribute_names:
            config = getattr(my_class, attribute_name)
            if attribute_name not in self._config:
                self._config[attribute_name] = config.default

            if config.required and self._config.get(attribute_name) is None:
                raise ValueError(
                    f"Missing required configuration property '{attribute_name}' for class '{my_class.__name__}'"
                )

        # loop through a second time to have the configs check their values
        # we do this as a separate step because we want to make sure required and default
        # values are specified before we have the configs do their validation.
        for attribute_name in attribute_names:
            getattr(my_class, attribute_name).finalize_and_validate_configuration(self)
            if attribute_name not in self._config:
                self._config[attribute_name] = None



from __future__ import annotations

import inspect
from typing import Any, Callable, Type

import clearskies.configs
import clearskies.exceptions
from clearskies import autodoc, typing
from clearskies.endpoint import Endpoint
from clearskies.functional import routing, string
from clearskies.input_outputs import InputOutput


class HealthCheck(Endpoint):
    """
    An endpoint that returns 200/500 to denote if backend services are functional.

    You can provide dependency injection names, classes, or a callable.  When invoked, this endpoint
    will build/call all of them and, as long as none raise any exceptions, return a 200.

    HealthCheck endpoints are always public and ignore authentication/authorization settings.

    If you don't provide any configuration to the endpoint, it will always succeed:

    ```python
    import clearskies

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.HealthCheck(),
    )
    wsgi()
    ```

    which when invoked:

    ```bash
    $ curl 'http://localhost:8080' | jq
    {
        "status": "success",
        "error": "",
        "data": {},
        "pagination": {},
        "input_errors": {}
    }
    ```

    This example demonstrates a failed healthcheck by requesting the cursor (which attempts to connect to the database).
    Since no database has been setup/configured, it always fails:

    ```python
    import clearskies

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.HealthCheck(
            dependency_injection_names=["cursor"],
        ),
    )
    wsgi()
    ```

    And when invoked returns:

    ```bash
    $ curl 'http://localhost:8080' | jq
    {
        "status": "failure",
        "error": "",
        "data": {},
        "pagination": {},
        "input_errors": {}
    }
    ```

    with a status code of 500.
    """

    """
    A list of dependency injection names that should be fetched when the healthcheck endpoint is invoked.

    If any exceptions are raised when building the dependency injection parameters, the health check will return
    failure.
    """
    dependency_injection_names = clearskies.configs.StringList(default=[])

    """
    A list of classes to build with the dependency injection system.

    The If any exceptions are raised when building the classes, then the healthcheck will return a failure.
    In the following example, since the class-to-build requests the cursor, and we don't have a reachable
    database configured,

    ```python
    import clearskies

    class MyClass:
        def __init__(self, cursor):
            pass

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.HealthCheck(
            classes_to_build=[MyClass],
        ),
    )
    wsgi()
    ```
    """
    classes_to_build = clearskies.configs.Any(default=[])

    """
    A list of callables to invoke.

    Your callables can request any dependency injection names.  If any exceptions are raised, the healthcheck will
    return a failure.  The return value from the function is ignored.  In this example we request the cursor from
    the dependency injection system, which will call the healthcheck to fail since we don't have a database setup
    and configured:

    ```python
    import clearskies

    def my_function(cursor):
        pass

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.HealthCheck(
            callables=[my_function],
        ),
    )
    wsgi()
    ```
    """
    callables = clearskies.configs.Any(default=[])

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        dependency_injection_names: list[str] = [],
        classes_to_build: list[type] = [],
        callables: list[Callable] = [],
        description: str = "",
        url: str = "",
        request_methods: list[str] = ["GET"],
    ):
        # we need to call the parent but don't have to pass along any of our kwargs.  They are all optional in our parent, and our parent class
        # just stores them in parameters, which we have already done.  However, the parent does do some extra initialization stuff that we need,
        # which is why we have to call the parent.
        super().__init__()

    def handle(self, input_output: InputOutput) -> Any:
        try:
            for name in self.dependency_injection_names:
                self.di.build_from_name(name)

            for class_to_build in self.classes_to_build:
                self.di.build_class(class_to_build)

            for thing_to_call in self.callables:
                self.di.call_function(thing_to_call)
        except:
            return self.failure(input_output)

        return self.success(input_output, {})

    def documentation(self) -> list[autodoc.request.Request]:
        output_schema = self.model_class
        nice_model = string.camel_case_to_words(output_schema.__name__)
        output_autodoc = (autodoc.schema.Object(self.auto_case_internal_column_name("data"), children=[]),)

        description = self.description if self.description else "Health Check"
        return [
            autodoc.request.Request(
                description,
                [
                    self.documentation_success_response(
                        output_autodoc,  # type: ignore
                        description=description,
                    ),
                ],
                relative_path=self.url,
                request_methods=self.request_methods,
            ),
        ]



import re


def match_route(expected_route, incoming_route, allow_partial=False) -> tuple[bool, dict[str, str]]:
    """
    Check if two routes match, and returns the routing data if so.

    A partial match happens when the beginning of the incoming route matches the expected route.  It's okay for the
    incoming route to be longer because the routing system is hierarchical, so a partial match at the beginning
    can work.  e.g.:

    Expected route: `/users`
    Incoming route: `/users/orders/5`

    But note that it must fully match all route segments, so this is never a match:

    Expected route: `/user`
    Incoming route: `/users/orders/5`
    """
    expected_route = expected_route.strip("/")
    incoming_route = incoming_route.strip("/")

    expected_parts = expected_route.split("/")
    incoming_parts = incoming_route.split("/")

    # quick check: if there are less parts in the incoming route than the expected route, then we can't possibly match
    if len(incoming_parts) < len(expected_parts):
        return (False, {})
    # ditto the opposite, if we can't do a partial match
    if len(expected_parts) < len(incoming_parts) and not allow_partial:
        return (False, {})

    # if we got this far then we will do a more complete match, so let's find any routing parameters
    routing_data = {}
    routing_parameters = extract_url_parameter_name_map(expected_route)
    # we want it backwards
    routing_parameters_by_index = {value: key for (key, value) in routing_parameters.items()}
    for index in range(len(expected_parts)):
        if index in routing_parameters_by_index:
            if not incoming_parts[index]:
                return (False, {})
            routing_data[routing_parameters_by_index[index]] = incoming_parts[index]
        else:
            if expected_parts[index] != incoming_parts[index]:
                return (False, {})

    return (True, routing_data)


def extract_url_parameter_name_map(url: str) -> dict[str, int]:
    """
    Create a map to help match URLs with routing parameters.

    Routing parameters are either brace enclosed or start with colons:

    ```python
    print(
        routing.extract_url_parameter_name_map("my/path/{some_parameter}/:other_parameter/more/paths")
    )
    # prints {"some_parameter": 2, "other_parameter": 3}
    ```

    Note that leading and trailing slashes are stripped, so "/my/path/{id}" and "my/path/{id}" give identical
    parameter maps: `{"id": 2}`
    """
    parameter_name_map = {}
    path_parts = url.strip("/").split("/")
    for index, part in enumerate(path_parts):
        if not part:
            continue
        if part[0] == ":":
            match = re.match("^:(\\w[\\w\\d_]{0,})$", part)
        else:
            if part[0] != "{":
                continue
            if part[-1] != "}":
                raise ValueError(
                    f"Invalid route configuration for URL '{url}': section '{part}'"
                    + " starts with a '{' but does not end with one"
                )
            match = re.match("^{(\\w[\\w\\d_]{0,})\\}$", part)
        if not match:
            raise ValueError(
                f"Invalid route configuration for URL '{url}', section '{part}': resource identifiers must start with a letter and contain only letters, numbers, and underscores"
            )
        parameter_name = match.group(1)
        if parameter_name in parameter_name_map:
            raise ValueError(
                f"Invalid route configuration for URL '{url}', a URL path named '{parameter_name}' appeared more than once."
            )
        parameter_name_map[parameter_name] = index
    return parameter_name_map



from typing import Any


class Response:
    response: Any = None
    formatted_schema: Any = None
    oai3_schema_resolver: Any = None
    status_code: Any = None

    def __init__(self, oai3_schema_resolver):
        self.oai3_schema_resolver = oai3_schema_resolver

    def set_response(self, response):
        self.response = response
        self.status_code = self.response.status
        self.formatted_schema = self.oai3_schema_resolver(response.schema)

    def convert(self):
        schema = {
            "description": self.response.description,
            "content": {
                "application/json": {
                    "schema": self.oai3_schema_resolver(self.response.schema).convert(),
                }
            },
        }

        return schema



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

import clearskies.di
import clearskies.typing
from clearskies import configs, parameters_to_properties  # type: ignore
from clearskies.columns.datetime import Datetime

if TYPE_CHECKING:
    from clearskies import Model


class Created(Datetime):
    """
    The created column records the time that a record is created.

    This will always populate the column when the model is first created.  If you attempt to set a value
    to this column on create then it will be overwritten.

    ```python
    import clearskies


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        created = clearskies.columns.Created()


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            lambda my_models: my_models.create({"name": "An Example"}),
            model_class=MyModel,
            readable_column_names=["id", "name", "created"],
        ),
        classes=[MyModel],
    )
    cli()
    ```

    And if you execute this you'll see that the `created` column was automatically populated:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "c54d74ac-5282-439e-af4f-23efb9ba96d4",
            "name": "An Example",
            "created": "2025-05-09T19:58:43+00:00",
        },
        "pagination": {},
        "input_errors": {},
    }
    ```
    """

    """
    Created fields are never writeable because they always set the created time automatically.
    """
    is_writeable = configs.Boolean(default=False)
    _descriptor_config_map = None

    now = clearskies.di.inject.Now()

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        date_format: str = "%Y-%m-%d %H:%M:%S",
        in_utc: bool = True,
        backend_default: str = "0000-00-00 00:00:00",
        is_readable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        pass

    def pre_save(self, data: dict[str, Any], model: Model) -> dict[str, Any]:
        if model:
            return data
        now = self.now
        if self.timezone_aware:
            now = now.astimezone(self.timezone)
        data = {**data, self.name: now}
        if self.on_change_pre_save:
            data = self.execute_actions_with_data(self.on_change_pre_save, model, data)
        return data



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Self

import clearskies.configurable
import clearskies.di
import clearskies.end
from clearskies import exceptions
from clearskies.authentication import Authentication, Authorization, Public
from clearskies.endpoint import Endpoint
from clearskies.functional import routing
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import SecurityHeader


class EndpointGroup(
    clearskies.end.End,  # type: ignore
    clearskies.configurable.Configurable,
    clearskies.di.InjectableProperties,
):
    """
    An endpoint group brings endpoints together: it basically handles routing.

    The endpoint group accepts a list of endpoints/endpoint groups and routes requests to them.  You can set a URL for
    the endpoint group, and this becomes a URL prefix for all of the endpoints under it.  Note that all routing is
    greedy, which means you want to put endpoints with more specific URLs first.  Here's an example of how
    you can use them to build a fully functional API that manages both users and companies.  Each individual
    endpoint is defined for the purpose of the example, but note that in practice you could accomplish this same
    thing with much less code by using the RestfulApi endpoint:

    ```python
    import clearskies
    from clearskies.validators import Required, Unique
    from clearskies import columns


    class Company(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = columns.Uuid()
        name = columns.String(
            validators=[
                Required(),
                Unique(),
            ]
        )


    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = columns.Uuid()
        name = columns.String(validators=[Required()])
        username = columns.String(
            validators=[
                Required(),
                Unique(),
            ]
        )
        age = columns.Integer(validators=[Required()])
        created_at = columns.Created()
        updated_at = columns.Updated()
        company_id = columns.BelongsToId(
            Company,
            readable_parent_columns=["id", "name"],
            validators=[Required()],
        )
        company = columns.BelongsToModel("company_id")


    readable_user_column_names = [
        "id",
        "name",
        "username",
        "age",
        "created_at",
        "updated_at",
        "company",
    ]
    writeable_user_column_names = ["name", "username", "age", "company_id"]
    users_api = clearskies.EndpointGroup(
        [
            clearskies.endpoints.Update(
                model_class=User,
                url="/:id",
                readable_column_names=readable_user_column_names,
                writeable_column_names=writeable_user_column_names,
            ),
            clearskies.endpoints.Delete(
                model_class=User,
                url="/:id",
            ),
            clearskies.endpoints.Get(
                model_class=User,
                url="/:id",
                readable_column_names=readable_user_column_names,
            ),
            clearskies.endpoints.Create(
                model_class=User,
                readable_column_names=readable_user_column_names,
                writeable_column_names=writeable_user_column_names,
            ),
            clearskies.endpoints.SimpleSearch(
                model_class=User,
                readable_column_names=readable_user_column_names,
                sortable_column_names=readable_user_column_names,
                searchable_column_names=readable_user_column_names,
                default_sort_column_name="name",
            ),
        ],
        url="users",
    )

    readable_company_column_names = ["id", "name"]
    writeable_company_column_names = ["name"]
    companies_api = clearskies.EndpointGroup(
        [
            clearskies.endpoints.Update(
                model_class=Company,
                url="/:id",
                readable_column_names=readable_company_column_names,
                writeable_column_names=writeable_company_column_names,
            ),
            clearskies.endpoints.Delete(
                model_class=Company,
                url="/:id",
            ),
            clearskies.endpoints.Get(
                model_class=Company,
                url="/:id",
                readable_column_names=readable_company_column_names,
            ),
            clearskies.endpoints.Create(
                model_class=Company,
                readable_column_names=readable_company_column_names,
                writeable_column_names=writeable_company_column_names,
            ),
            clearskies.endpoints.SimpleSearch(
                model_class=Company,
                readable_column_names=readable_company_column_names,
                sortable_column_names=readable_company_column_names,
                searchable_column_names=readable_company_column_names,
                default_sort_column_name="name",
            ),
        ],
        url="companies",
    )

    wsgi = clearskies.contexts.WsgiRef(clearskies.EndpointGroup([users_api, companies_api]))
    wsgi()
    ```

    Usage then works exactly as expected:

    ```bash
    $ curl 'http://localhost:8080/companies' -d '{"name": "Box Store"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "f073ee4d-318d-4e0b-a796-f450c40aa771",
            "name": "Box Store"
        },
        "pagination": {},
        "input_errors": {}
    }

    curl 'http://localhost:8080/users' -d '{"name": "Bob Brown", "username": "bobbrown", "age": 25, "company_id": "f073ee4d-318d-4e0b-a796-f450c40aa771"}'
    curl 'http://localhost:8080/users' -d '{"name": "Jane Doe", "username": "janedoe", "age": 32, "company_id": "f073ee4d-318d-4e0b-a796-f450c40aa771"}'

    $ curl 'http://localhost:8080/users' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "68cbb9e9-689a-4ae0-af77-d60e4cb344f1",
                "name": "Bob Brown",
                "username": "bobbrown",
                "age": 25,
                "created_at": "2025-06-08T10:40:37+00:00",
                "updated_at": "2025-06-08T10:40:37+00:00",
                "company": {
                    "id": "f073ee4d-318d-4e0b-a796-f450c40aa771",
                    "name": "Box Store"
                }
            },
            {
                "id": "e69c4ebf-38b1-40d2-b523-5d58f5befc7b",
                "name": "Jane Doe",
                "username": "janedoe",
                "age": 32,
                "created_at": "2025-06-08T10:41:04+00:00",
                "updated_at": "2025-06-08T10:41:04+00:00",
                "company": {
                    "id": "f073ee4d-318d-4e0b-a796-f450c40aa771",
                    "name": "Box Store"
                }
            }
        ],
        "pagination": {
            "number_results": 2,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }

    ```
    """

    """
    The dependency injection container
    """
    di = clearskies.di.inject.Di()
    url = clearskies.configs.String(default="")
    response_headers = clearskies.configs.StringListOrCallable(default=[])
    authentication = clearskies.configs.Authentication(default=Public())
    authorization = clearskies.configs.Authorization(default=Authorization())
    security_headers = clearskies.configs.SecurityHeaders(default=[])
    cors_header: SecurityHeader = None  # type: ignore
    has_cors: bool = False
    endpoints_initialized = False

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        endpoints: list[Endpoint | Self],
        url: str = "",
        response_headers: list[str | Callable[..., list[str]]] = [],
        security_headers: list[SecurityHeader] = [],
        authentication: Authentication = Public(),
        authorization: Authorization = Authorization(),
    ):
        self.finalize_and_validate_configuration()
        for security_header in self.security_headers:
            if not security_header.is_cors:
                continue
            self.cors_header = security_header
            self.has_cors = True
            break

        if not endpoints:
            raise ValueError(
                "An endpoint group must receive a list of endpoints/endpoint groups, but my list of endpoints is empty."
            )
        if not isinstance(endpoints, list):
            raise ValueError(
                f"An endpoint group must receive a list of endpoints/endpoint groups, but instead of a list I found an object of type '{endpoints.__class__.__name__}'"
            )
        for index, endpoint in enumerate(endpoints):
            if not isinstance(endpoint, Endpoint) and not isinstance(endpoint, self.__class__):
                raise ValueError(
                    f"An endpoint group must receive a list of endpoints/endpoint groups, but item #{index + 1} was neither an endpoint nor an endpoint group, but an object of type '{endpoints.__class__.__name__}'"
                )
            if self.url.strip("/"):
                endpoint.add_url_prefix(self.url)

    def matches_request(self, input_output: InputOutput, allow_partial=True) -> bool:
        """Whether or not we can handle an incoming request based on URL and request method."""
        expected_url = self.url.strip("/")
        incoming_url = input_output.get_full_path().strip("/")
        if not expected_url and not incoming_url:
            return True
        (matches, routing_data) = routing.match_route(expected_url, incoming_url, allow_partial=allow_partial)
        return matches

    def populate_routing_data(self, input_output: InputOutput) -> Any:
        # only endpoints (not the endpoint group) can handle this because the endpoint group doesn't have the full url
        return None

    def handle(self, input_output):
        if not self.endpoints_initialized:
            self.endpoints_initialized = True
            for endpoint in self.endpoints:
                endpoint.injectable_properties(self.di)

        has_match = False
        for endpoint in self.endpoints:
            if not endpoint.matches_request(input_output):
                continue
            has_match = True
            break

        if not has_match:
            return self.error(input_output, "Not Found", 404)

        self.add_response_headers(input_output)
        return endpoint(input_output)

    def error(self, input_output: InputOutput, message: str, status_code: int) -> Any:
        """Return a client-side error (e.g. 400)."""
        return self.respond_json(input_output, {"status": "client_error", "error": message}, status_code)



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

import dateparser

import clearskies.configs
import clearskies.parameters_to_properties
from clearskies.validator import Validator

if TYPE_CHECKING:
    import clearskies.model


class AfterColumn(Validator):
    """The name of the other date column for comparison."""

    other_column_name = clearskies.configs.String(default="", required=True)

    """
    If true, then this column is allowed to be eqaul to the other column.
    """
    allow_equal = clearskies.configs.Boolean(default=False)

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(self, other_column_name: str, allow_equal: bool = False):
        self.other_column_name = other_column_name
        self.allow_equal = allow_equal
        self.finalize_and_validate_configuration()

    def check(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        # we won't check anything for missing values (columns should be required if that is an issue)
        if not data.get(column_name):
            return ""
        my_value = data[column_name]
        other_value = data.get(self.other_column_name, getattr(model, self.other_column_name))
        # again, no checks for non-values
        if not other_value:
            return ""

        my_value = dateparser.parse(my_value) if isinstance(my_value, str) else my_value
        if not my_value:
            return f"'{column_name}' was not a valid date."

        if type(other_value) != str and type(other_value) != datetime.datetime:
            return f"'{self.other_column_name}' was not a valid date."
        other_value = dateparser.parse(other_value) if isinstance(other_value, str) else other_value
        if not other_value:
            return f"'{self.other_column_name}' was not a valid date."

        return self.date_comparison(my_value, other_value, column_name)

    def date_comparison(
        self, incoming_date: datetime.datetime, comparison_date: datetime.datetime, column_name: str
    ) -> str:
        if incoming_date == comparison_date:
            return "" if self.allow_equal else f"'{column_name}' must be after '{self.other_column_name}'"

        if incoming_date < comparison_date:
            return f"'{column_name}' must be after '{self.other_column_name}'"
        return ""



from typing import Any, Callable

import clearskies
from clearskies import parameters_to_properties
from clearskies.autodoc.schema import Integer as AutoDocInteger
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.autodoc.schema import String as AutoDocString
from clearskies.backends.backend import Backend
from clearskies.di import InjectableProperties, inject
from clearskies.functional import routing, string


class SecretsBackend(Backend):
    """
    Fetch and store data from a secret provider.

    ## Installing Dependencies

    Clearskies uses Akeyless by default to manage the secrets.
    This is not installed by default, but is a named extra that you can install when needed via:

    ```bash
    pip install clear-skies[secrets]
    ```
    """

    """The secrets instance."""
    secrets = inject.Secrets()

    can_count: bool = False

    def __init__(self):
        pass

    def check_query(self, query: clearskies.query.Query) -> None:
        if not query.conditions:
            raise KeyError(f"You must search by an id when using the secrets backend.")

    def update(self, id: str, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:  # type: ignore[override]
        """Update the record with the given id with the information from the data dictionary."""
        folder_path = self._make_folder_path(model, id)
        for key, value in data.items():
            if key == model.id_column_name:
                continue
            self.secrets.update(f"{folder_path}{key}", value)

        # and now query again to fetch the updated record.
        return self.records(
            clearskies.query.Query(
                model.__class__, conditions=[clearskies.query.Condition(f"{model.id_column_name}={id}")]
            )
        )[0]

    def create(self, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        if not model.id_column_name in data:
            raise ValueError(
                f"You must provide '{model.id_column_name}' when creating a record with the secrets backend"
            )
        return self.update(data[model.id_column_name], data, model)

    def delete(self, id: str, model: clearskies.model.Model) -> bool:  # type: ignore[override]
        """
        Delete the record with the given id.

        Note that this isn't implemented yet, and always returns True.
        """
        return True

    def records(
        self, query: clearskies.query.Query, next_page_data: dict[str, str | int] | None = None
    ) -> list[dict[str, Any]]:
        """Return a list of records that match the given query configuration."""
        self.check_query(query)
        for condition in query.conditions:
            if condition.operator != "=":
                raise ValueError(
                    f"I'm not very smart and only know how to search with the equals operator, but I received a condition of {condition.parsed}.  If you need to support this, you'll have to extend the ApiBackend and overwrite the build_records_request method."
                )
            if condition.column_name == query.model_class.id_column_name:
                id = condition.values[0]
                break
        if id is None:
            raise ValueError(f"You must search by '{query.model_class.id_column_name}' when using the secrets backend")

        folder_path = self._make_folder_path(query.model_class, id)
        data = {query.model_class.id_column_name: id}
        for path in self.secrets.list_secrets(folder_path):
            data[path[len(folder_path) :]] = self.secrets.get(path)
        return [data]

    def _make_folder_path(self, model, id):
        return model.table_name().rstrip("/") + "/" + id.strip("/") + "/"

    def validate_pagination_data(self, data: dict[str, Any], case_mapping: Callable[[str], str]) -> str:
        return ""

    def allowed_pagination_keys(self) -> list[str]:
        return []

    def documentation_pagination_next_page_response(self, case_mapping: Callable) -> list[Any]:
        return []

    def documentation_pagination_parameters(self, case_mapping: Callable) -> list[tuple[AutoDocSchema, str]]:
        return []

    def documentation_pagination_next_page_example(self, case_mapping: Callable) -> dict[str, Any]:
        return {}



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.autodoc.schema import Object as AutoDocObject
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.autodoc.schema import String as AutoDocString
from clearskies.columns.string import String
from clearskies.di.inject import InputOutput
from clearskies.functional import validations

if TYPE_CHECKING:
    from clearskies import Column, Model


class BelongsToId(String):
    """
    Declares that this model belongs to another - that it has a parent.

    ## Usage

    The way that a belongs to relationship works is that the child model (e.g. the one with
    the BelongsToId column) needs to have a column that stores the id of the parent it is related
    to.  Then you can attach the BelongsToModel class and point it to the column containing the
    id.  If you allow the end-user to set the parent id in a save action, the belongs to column
    will automatically verify that the given id corresponds to an actual record.  Here's a simple
    usage example:

    ```python
    import clearskies


    class Category(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()


    class Product(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        category_id = clearskies.columns.BelongsToId(Category)
        category = clearskies.columns.BelongsToModel("category_id")


    def test_belongs_to(products: Product, categories: Category):
        toys = categories.create({"name": "Toys"})
        auto = categories.create({"name": "Auto"})

        # Note: we set the cateogry by setting "category_id"
        ball = products.create({"name": "ball", "category_id": toys.id})

        # note: we set the category by saving a category model to "category"
        fidget_spinner = products.create({"name": "Fidget Spinner", "category": toys})

        return {
            "ball_category": ball.category.name,
            "fidget_spinner_category": fidget_spinner.category.name,
            "ball_id_check": ball.category_id == ball.category.id,
            "ball_fidget_id_check": fidget_spinner.category_id == ball.category.id,
        }


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(test_belongs_to),
        classes=[Category, Product],
    )

    if __name__ == "__main__":
        cli()
    ```

    ## Circular Dependency Trees

    The opposite of a BelongsToId relationship is a HasMany relationship.  It's common
    for the child model to contain a BelonsToId column to point to the parent, and then
    have the parent contain a HasMany column to point to the child.  This creates circular
    depenency errors in python.  To work around this, clearskies requires the addition of
    a "model reference" class that looks like this:

    ```python
    import some_model


    class SomeModelReference:
        def get_model_class(self):
            return some_model.SomeModel
    ```

    These have to live in their own file, should use relative imports to import the file containing
    the model, and should not be imported into the module they live in.  So, sticking with the example
    of categories and products, you would have the following directory structure:

    ```
    ├── models
    │   ├── category.py
    │   ├── category_reference.py
    │   ├── product.py
    │   └── product_reference.py
    │
    └── app.py
    ```

    The files would then contain:

    category.py
    ```python
    import clearskies
    import models.product_reference


    class Category(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        products = clearskies.columns.HasMany(product_reference.ProductReference)
    ```

    category_reference.py
    ```python
    from clearskies.model import ModelClassReference
    from . import cateogry


    class CategoryReference(ModelClassReference):
        def get_model_class(self):
            return category.Category
    ```

    product.py
    ```python
    import clearskies
    import models.category_reference


    class Product(clearskies.model.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        category_id = clearskies.columns.BelongsToId(CategoryReference)
        category = clearskies.columns.BelongsToModel("category_id")
    ```

    product_reference.py
    ```python
    from clearskies.model import ModelClassReference
    from . import product


    class ProductReference(ModelClassReference):
        def get_model_class(self):
            return product.Product
    ```
    """

    """ The model class we belong to. """
    parent_model_class = configs.ModelClass(required=True)

    """
    The name of the property used to fetch the parent model itself.

    Note that this isn't set explicitly, but by adding a BelongsToModel column to the model.
    """
    model_column_name = configs.String()

    """
    The list of columns from the parent that should be included when converting this column to JSON.

    When configuring readable columns for an endpoint, you can specify the BelongsToModel column.
    If you do this, you must set readable_parent_columns on the BelongsToId column to specify which
    columns from the parent model should be returned in the response.  See this example:

    ```python
    import clearskies

    class Owner(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()

    class Pet(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        owner_id = clearskies.columns.BelongsToId(
            Owner,
            readable_parent_columns=["id", "name"],
        )
        owner = clearskies.columns.BelongsToModel("owner_id")

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.List(
            Pet,
            sortable_column_names=["id", "name"],
            readable_column_names=["id", "name", "owner"],
            default_sort_column_name="name",
        ),
        classes=[Owner, Pet],
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": Owner,
                    "records": [
                        {"id": "1-2-3-4", "name": "John Doe"},
                        {"id": "5-6-7-8", "name": "Jane Doe"},
                    ],
                },
                {
                    "model_class": Pet,
                    "records": [
                        {"id": "a-b-c-d", "name": "Fido", "owner_id": "1-2-3-4"},
                        {"id": "e-f-g-h", "name": "Spot", "owner_id": "1-2-3-4"},
                        {"id": "i-j-k-l", "name": "Puss in Boots", "owner_id": "5-6-7-8"},
                    ],
                },
            ],
        }
    )

    if __name__ == "__main__":
        cli()
    ```

    With readable_parent_columns set in the Pet.owner_id column, and owner set in the list configuration,
    The owner id and name are included in the `owner` key of the returned Pet dictionary:

    ```bash
    $ ./test.py  | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "a-b-c-d",
                "name": "Fido",
                "owner": {
                    "id": "1-2-3-4",
                    "name": "John Doe"
                }
            },
            {
                "id": "i-j-k-l",
                "name": "Puss in Boots",
                "owner": {
                    "id": "5-6-7-8",
                    "name": "Jane Doe"
                }
            },
            {
                "id": "e-f-g-h",
                "name": "Spot",
                "owner": {
                    "id": "1-2-3-4",
                    "name": "John Doe"
                }
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```

    """
    readable_parent_columns = configs.ReadableModelColumns("parent_model_class")

    """
    The type of join to use when searching on the parent.
    """
    join_type = configs.Select(["LEFT", "INNER", "RIGHT"], default="LEFT")

    """
    Any additional conditions to place on the parent table when finding related records.

    where should be a list containing a combination of conditions-as-strings, queries built from the columns
    themselves, or callable functions which accept the model and apply filters.  This is primarily used in
    input validation to exclude values as allowed parents.
    """
    where = configs.Conditions()

    input_output = InputOutput()
    wants_n_plus_one = True
    _allowed_search_operators = ["="]
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        parent_model_class,
        readable_parent_columns: list[str] = [],
        join_type: str | None = None,
        where: clearskies.typing.condition | list[clearskies.typing.condition] = [],
        default: str | None = None,
        setable: str | Callable | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    @property
    def parent_model(self) -> Model:
        parents = self.di.build(self.parent_model_class, cache=True)
        if not self.where:
            return parents

        return self.apply_wheres(parents)

    def apply_wheres(self, parents: Model) -> Model:
        if not self.where:
            return parents

        for index, where in enumerate(self.where):
            if callable(where):
                parents = self.di.call_function(where, model=parents, **self.input_output.get_context_for_callables())
                if not validations.is_model(parents):
                    raise ValueError(
                        f"Configuration error for {self.model_class.__name__}.{self.name}: when 'where' is a callable, it must return a model class, but when the callable in where entry #{index + 1} was called, it returned something else."
                    )
            else:
                parents = parents.where(where)
        return parents

    @property
    def parent_columns(self) -> dict[str, Any]:
        return self.parent_model_class.get_columns()

    def input_error_for_value(self, value: str, operator: str | None = None) -> str:
        parent_check = super().input_error_for_value(value)
        if parent_check:
            return parent_check
        parent_model = self.parent_model
        matching_parents = parent_model.where(f"{parent_model.id_column_name}={value}")
        matching_parents = self.apply_wheres(matching_parents)
        matching_parents = matching_parents.where_for_request(
            matching_parents,
            self.input_output.routing_data,
            self.input_output.authorization_data,
            self.input_output,
        )
        if not len(matching_parents):
            return f"Invalid selection for {self.name}: record does not exist"
        return ""

    def n_plus_one_add_joins(self, model: Model, column_names: list[str] = []) -> Model:
        """Add any additional joins to solve the N+1 problem."""
        if not column_names:
            column_names = self.readable_parent_columns
        if not column_names:
            return model

        model = self.add_join(model)
        alias = self.join_table_alias()
        parent_id_column_name = self.parent_model.id_column_name
        select_parts = [f"{alias}.{column_name} AS {alias}_{column_name}" for column_name in column_names]
        if parent_id_column_name not in column_names:
            select_parts.append(f"{alias}.{parent_id_column_name} AS {alias}_{parent_id_column_name}")
        return model.select(", ".join(select_parts))

    def add_join(self, model: Model) -> Model:
        parent_table = self.parent_model.destination_name()
        alias = self.join_table_alias()

        if model.is_joined(parent_table, alias=alias):
            return model

        join_type = "LEFT " if self.join_type == "LEFT" else ""
        own_table_name = model.destination_name()
        parent_id_column_name = self.parent_model.id_column_name
        return model.join(
            f"{join_type}JOIN {parent_table} as {alias} on {alias}.{parent_id_column_name}={own_table_name}.{self.name}"
        )

    def join_table_alias(self) -> str:
        return self.parent_model.destination_name() + "_" + self.name

    def is_allowed_operator(self, operator, relationship_reference=None):
        """Proces user data to decide if the end-user is specifying an allowed operator."""
        if not relationship_reference:
            return "="
        parent_columns = self.parent_columns
        if relationship_reference not in self.parent_columns:
            raise ValueError(
                "I was asked to search on a related column that doens't exist.  This shouldn't have happened :("
            )
        return self.parent_columns[relationship_reference].is_allowed_operator(operator)

    def check_search_value(self, value, operator=None, relationship_reference=None):
        if not relationship_reference:
            return self.input_error_for_value(value, operator=operator)
        parent_columns = self.parent_columns
        if relationship_reference not in self.parent_columns:
            raise ValueError(
                "I was asked to search on a related column that doens't exist.  This shouldn't have happened :("
            )
        return self.parent_columns[relationship_reference].check_search_value(value, operator=operator)

    def is_allowed_search_operator(self, operator: str, relationship_reference: str = "") -> bool:
        if not relationship_reference:
            return operator in self._allowed_search_operators
        parent_columns = self.parent_columns
        if relationship_reference not in self.parent_columns:
            raise ValueError(
                "I was asked to check search operators on a related column that doens't exist.  This shouldn't have happened :("
            )
        return self.parent_columns[relationship_reference].is_allowed_search_operator(
            operator, relationship_reference=relationship_reference
        )

    def allowed_search_operators(self, relationship_reference: str = ""):
        if not relationship_reference:
            return self._allowed_search_operators
        parent_columns = self.parent_columns
        if relationship_reference not in self.parent_columns:
            raise ValueError(
                "I was asked for allowed search operators on a related column that doens't exist.  This shouldn't have happened :("
            )
        return self.parent_columns[relationship_reference].allowed_search_operators()

    def add_search(
        self, model: clearskies.model.Model, value: str, operator: str = "", relationship_reference: str = ""
    ) -> clearskies.model.Model:
        if not relationship_reference:
            return super().add_search(model, value, operator=operator)

        if relationship_reference not in self.parent_columns:
            raise ValueError(
                "I was asked to search on a related column that doens't exist.  This shouldn't have happened :("
            )

        model = self.add_join(model)
        related_column = self.parent_columns[relationship_reference]
        alias = self.join_table_alias()
        return model.where(related_column.build_condition(value, operator=operator, column_prefix=f"{alias}."))

    def documentation(
        self, name: str | None = None, example: str | None = None, value: str | None = None
    ) -> list[AutoDocSchema]:
        columns = self.parent_columns
        parent_id_column_name = self.parent_model.id_column_name
        parent_properties = [columns[parent_id_column_name].documentation()]
        parent_id_doc = AutoDocString(name if name is not None else self.name)

        readable_parent_columns = self.readable_parent_columns
        if not readable_parent_columns:
            return [parent_id_doc]

        for column_name in readable_parent_columns:
            if column_name == parent_id_column_name:
                continue
            parent_properties.append(columns[column_name].documentation())

        return [
            parent_id_doc,
            AutoDocObject(
                self.model_column_name,
                parent_properties,
            ),
        ]



from .mysql_connection_dynamic_producer import MySQLConnectionDynamicProducer
from .mysql_connection_dynamic_producer_via_ssh_cert_bastion import MySQLConnectionDynamicProducerViaSSHCertBastion


def mysql_connection_dynamic_producer(producer_name=None, database_host=None, database_name=None):
    return MySQLConnectionDynamicProducer(
        producer_name=producer_name,
        database_host=database_host,
        database_name=database_name,
    )


def mysql_connection_dynamic_producer_via_ssh_cert_bastion(
    producer_name=None,
    bastion_host=None,
    bastion_username=None,
    public_key_file_path=None,
    cert_issuer_name=None,
    database_host=None,
    database_name=None,
    local_proxy_port=None,
):
    return MySQLConnectionDynamicProducerViaSSHCertBastion(
        producer_name=producer_name,
        bastion_host=bastion_host,
        bastion_username=bastion_username,
        cert_issuer_name=cert_issuer_name,
        public_key_file_path=public_key_file_path,
        database_host=database_host,
        database_name=database_name,
        local_proxy_port=local_proxy_port,
    )



from .additional_config import AdditionalConfig


class AdditionalConfigAutoImport(AdditionalConfig):
    """
    AdditionalConfig, but now with auto import.

    This works exactly like the AdditionalConfig class, but will be automatically imported into the Di
    container if found in a module being imported into the Di container.  This just provides a way for
    modules to easily inject names into the Di system.  In order to be found, an imported module
    just needs to make sure that it imports the class that extends the AdditionalConfigAutoImport class.

    Note that automatically-imported AdditionalConfig classes automatically have lower priority than
    any AdditionalConfig classes added explicitly to the Di container.
    """

    pass



import clearskies.configs
import clearskies.parameters_to_properties
from clearskies.security_header import SecurityHeader


class Csp(SecurityHeader):
    header_name = "content-security-policy"
    default_src = clearskies.configs.String()
    script_src = clearskies.configs.String()
    style_src = clearskies.configs.String()
    img_src = clearskies.configs.String()
    connect_src = clearskies.configs.String()
    font_src = clearskies.configs.String()
    object_src = clearskies.configs.String()
    media_src = clearskies.configs.String()
    frame_src = clearskies.configs.String()
    sandbox = clearskies.configs.String()
    report_uri = clearskies.configs.String()
    child_src = clearskies.configs.String()
    form_action = clearskies.configs.String()
    frame_ancestors = clearskies.configs.String()
    plugin_types = clearskies.configs.String()
    base_uri = clearskies.configs.String()
    report_to = clearskies.configs.String()
    worker_src = clearskies.configs.String()
    manifest_src = clearskies.configs.String()
    prefetch_src = clearskies.configs.String()
    navigate_to = clearskies.configs.String()

    directives = [
        "default_src",
        "script_src",
        "style_src",
        "img_src",
        "connect_src",
        "font_src",
        "object_src",
        "media_src",
        "frame_src",
        "sandbox",
        "report_uri",
        "child_src",
        "form_action",
        "frame_ancestors",
        "plugin_types",
        "base_uri",
        "report_to",
        "worker_src",
        "manifest_src",
        "prefetch_src",
        "navigate_to",
    ]

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        default_src: str = "",
        script_src: str = "",
        style_src: str = "",
        img_src: str = "",
        connect_src: str = "",
        font_src: str = "",
        object_src: str = "",
        media_src: str = "",
        frame_src: str = "",
        sandbox: str = "",
        report_uri: str = "",
        child_src: str = "",
        form_action: str = "",
        frame_ancestors: str = "",
        plugin_types: str = "",
        base_uri: str = "",
        report_to: str = "",
        worker_src: str = "",
        manifest_src: str = "",
        prefetch_src: str = "",
        navigate_to: str = "",
    ):
        self.finalize_and_validate_configuration()

    def set_headers_for_input_output(self, input_output):
        parts = []
        for variable_name in self.directives:
            value = getattr(self, variable_name)
            if not value:
                continue
            if value.lower().strip() == "self":
                value = "'self'"
            header_key_name = variable_name.replace("_", "-")
            parts.append(f"{header_key_name} {value}")
        if not parts:
            return
        header_value = "; ".join(parts)
        input_output.response_headers.add(self.header_name, header_value)



class Sort:
    """Stores a sort directive."""

    """
    The name of the table to sort on.
    """
    table_name: str = ""

    """
    The name of the column to sort on.
    """
    column_name: str = ""

    """
    The direction to sort.
    """
    direction: str = ""

    def __init__(self, table_name: str, column_name: str, direction: str):
        if not column_name:
            raise ValueError("Missing 'column_name' for sort")
        direction = direction.upper().strip()
        if direction != "ASC" and direction != "DESC":
            raise ValueError(f"Invalid sort direction: should be ASC or DESC, not '{direction}'")
        self.table_name = table_name
        self.column_name = column_name
        self.direction = direction



from __future__ import annotations

import datetime
from collections import OrderedDict
from typing import TYPE_CHECKING, Any

import dateparser

import clearskies.configs
import clearskies.di
import clearskies.parameters_to_properties
from clearskies.validator import Validator

if TYPE_CHECKING:
    import clearskies.model


class InThePast(Validator, clearskies.di.InjectableProperties):
    utcnow = clearskies.di.inject.Utcnow()

    def check(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        if not data.get(column_name):
            return ""

        as_date = dateparser.parse(data[column_name]) if isinstance(data[column_name], str) else data[column_name]
        if not as_date:
            return f"'{column_name}' was not a valid date"
        if as_date.tzinfo == None:
            as_date = as_date.replace(tzinfo=datetime.timezone.utc)
        if as_date >= self.utcnow:
            return f"'{column_name}' must be in the past"
        return ""



from clearskies.configs import string


class Select(string.String):
    def __init__(self, allowed_values: list[str], required=False, default=None):
        self.allowed_values = allowed_values
        self.required = required
        self.default = default

    def __set__(self, instance, value: str):
        if value is None:
            return

        if not isinstance(value, str):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a string parameter"
            )
        if value not in self.allowed_values:
            error_prefix = self._error_prefix(instance)
            raise ValueError(
                f"{error_prefix} attempt to set a value of type '{value}' which is not in the list of allowed values.  It must be one of '"  # type: ignore
                + "', '".join(self.allowed_values)
                + "'"
            )
        instance._set_config(self, value)



from typing import Any

from clearskies.di.injectable import Injectable


class ByName(Injectable):
    def __init__(self, name: str, cache: bool = True):
        if not isinstance(name, str):
            raise TypeError(
                f"I expected a string for the first argument to clearskies.di.inject.ByName, but I received an object of type '{name.__class__.__name__}' instead."
            )
        self.name = name
        self.cache = cache

    def __get__(self, instance, parent) -> Any:
        if instance is None:
            return self  # type: ignore
        return self._di.build_from_name(self.name, cache=self.cache)



from __future__ import annotations

import datetime
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable

import clearskies.endpoint
import clearskies.endpoint_group
from clearskies.di import Di
from clearskies.di.additional_config import AdditionalConfig
from clearskies.input_outputs import Programmatic

if TYPE_CHECKING:
    from clearskies.input_outputs import InputOutput


class Context:
    di: Di = None  # type: ignore

    def __init__(
        self,
        application: Callable | clearskies.endpoint.Endpoint | clearskies.endpoint_group.EndpointGroup,
        classes: type | list[type] = [],
        modules: ModuleType | list[ModuleType] = [],
        bindings: dict[str, Any] = {},
        additional_configs: AdditionalConfig | list[AdditionalConfig] = [],
        class_overrides: dict[type, Any] = {},
        overrides: dict[str, type] = {},
        now: datetime.datetime | None = None,
        utcnow: datetime.datetime | None = None,
    ):
        self.di = Di(
            classes=classes,
            modules=modules,
            bindings=bindings,
            additional_configs=additional_configs,
            class_overrides=class_overrides,
            now=now,
            utcnow=utcnow,
        )
        self.application = application

    def execute_application(self, input_output: InputOutput):
        if hasattr(self.application, "injectable_properties"):
            self.application.injectable_properties(self.di)
            return self.application(input_output)
        elif callable(self.application):
            try:
                return input_output.respond(
                    self.di.call_function(self.application, **input_output.get_context_for_callables())
                )
            except clearskies.exceptions.ClientError as e:
                return input_output.respond(str(e), 400)
            except clearskies.exceptions.Authentication as e:
                return input_output.respond(str(e), 401)
            except clearskies.exceptions.Authorization as e:
                return input_output.respond(str(e), 403)
            except clearskies.exceptions.NotFound as e:
                return input_output.respond(str(e), 404)
            except clearskies.exceptions.MovedPermanently as e:
                return input_output.respond(str(e), 302)
            except clearskies.exceptions.MovedTemporarily as e:
                return input_output.respond(str(e), 307)

    def __call__(
        self,
        url: str = "",
        request_method: str = "GET",
        body: str | dict[str, Any] | list[Any] = "",
        query_parameters: dict[str, str] = {},
        request_headers: dict[str, str] = {},
    ):
        return self.execute_application(
            Programmatic(
                url=url,
                request_method=request_method,
                body=body,
                query_parameters=query_parameters,
                request_headers=request_headers,
            )
        )

    def build(self, thing: Any, cache: bool = False) -> Any:
        return self.di.build(thing, cache=cache)



from clearskies.configs.string import String
from clearskies.functional import string


class ModelDestinationName(String):
    def __init__(self, model_column_config_name, required=False, default=None):
        self.required = required
        self.default = default
        self.model_column_config_name = model_column_config_name

    def get_model_class(self, instance):
        if self.model_column_config_name:
            return getattr(instance, self.model_column_config_name)
        return None

    def finalize_and_validate_configuration(self, instance):
        # we use the model class itself to decide on our value.  However,
        # if we don't have it yet, don't worry.  Someone will eventually tell us
        # what we need to know
        model_class = self.get_model_class(instance)
        if not model_class:
            return

        instance._set_config(self, model_class.destination_name())
        super().finalize_and_validate_configuration(instance)



from __future__ import annotations

from typing import TYPE_CHECKING

from clearskies.configs import config

if TYPE_CHECKING:
    from clearskies.schema import Schema as SchemaType


class Schema(config.Config):
    def __set__(self, instance, value: type[SchemaType]) -> None:
        if not hasattr(value, "get_columns"):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to parameter that requires a Schema."
            )
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> type[SchemaType]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



import clearskies.configs
import clearskies.parameters_to_properties
from clearskies.security_header import SecurityHeader


class Hsts(SecurityHeader):
    max_age = clearskies.configs.Integer(default=31536000)
    include_sub_domains = clearskies.configs.Boolean()

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        max_age: int = 31536000,
        include_sub_domains: bool = False,
    ):
        self.finalize_and_validate_configuration()

    def set_headers_for_input_output(self, input_output):
        value = f"max-age={self.max_age} ;"
        if self.include_sub_domains:
            value += " includeSubDomains"
        input_output.response_headers.add("strict-transport-security", value)



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

import clearskies.configs
from clearskies import parameters_to_properties
from clearskies.validator import Validator

if TYPE_CHECKING:
    import clearskies.model


class Unique(Validator):
    is_unique = True

    def check(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        # Unique is mildly tricky.  We obviously want to search the backend for the new value,
        # but we need to first skip this if our column is not being set, or if we're editing
        # the model and nothing is changing.
        if column_name not in data:
            return ""
        new_value = data[column_name]
        if model and getattr(model, column_name) == new_value:
            return ""

        as_query = model.as_query()
        matching_model = as_query.find(f"{column_name}={new_value}")
        if matching_model:
            return f"Invalid value for '{column_name}': the given value already exists, and must be unique."
        return ""



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Callable, Self, overload

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.column import Column

if TYPE_CHECKING:
    from clearskies import Model


class Json(Column):
    """
    A column to store generic data.

    ```python
    import clearskies


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        my_data = clearskies.columns.Json()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            MyModel,
            writeable_column_names=["my_data"],
            readable_column_names=["id", "my_data"],
        ),
        classes=[MyModel],
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080' -d '{"my_data":{"count":[1,2,3,4,{"thing":true}]}}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "63cbd5e7-a198-4424-bd35-3890075a2a5e",
            "my_data": {
                "count": [
                    1,
                    2,
                    3,
                    4,
                    {
                        "thing": true
                    }
                ]
            }
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    Note that there is no attempt to check the shape of the input passed into a JSON column.

    """

    is_searchable = configs.Boolean(default=False)
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        default: dict[str, Any] | None = None,
        setable: dict[str, Any] | Callable[..., dict[str, Any]] | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> dict[str, Any]:
        pass

    def __get__(self, instance, cls):
        return super().__get__(instance, cls)

    def __set__(self, instance, value: dict[str, Any]) -> None:
        instance._next_data[self.name] = value

    def from_backend(self, value) -> dict[str, Any] | list[Any] | None:
        if type(value) == list or type(value) == dict:
            return value
        if not value:
            return None
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return None

    def to_backend(self, data):
        if self.name not in data or data[self.name] is None:
            return data

        value = data[self.name]
        return {**data, self.name: value if isinstance(value, str) else json.dumps(value)}



import os
import socket
import subprocess
import time
from pathlib import Path

import clearskies.di


class MySQLConnectionDynamicProducerViaSSHCertBastion(clearskies.di.additional_config.AdditionalConfig):
    _config = None

    def __init__(
        self,
        producer_name=None,
        bastion_host=None,
        bastion_username=None,
        public_key_file_path=None,
        local_proxy_port=None,
        cert_issuer_name=None,
        database_host=None,
        database_name=None,
    ):
        # not using kwargs because I want the argument list to be explicit
        self.config = {
            "producer_name": producer_name,
            "bastion_host": bastion_host,
            "bastion_username": bastion_username,
            "public_key_file_path": public_key_file_path,
            "local_proxy_port": local_proxy_port,
            "cert_issuer_name": cert_issuer_name,
            "database_host": database_host,
            "database_name": database_name,
        }

    def provide_connection_details(self, environment, secrets):
        if not secrets:
            raise ValueError(
                "I was asked to connect to a database via an AKeyless dynamic producer but AKeyless itself wasn't configured.  Try setting the AKeyless auth method via clearskies.secrets.akeyless_[jwt|saml|aws_iam]_auth()"
            )

        home = str(Path.home())
        default_public_key_file_path = f"{home}/.ssh/id_rsa.pub"

        producer_name = self._fetch_config(environment, "producer_name", "akeyless_mysql_dynamic_producer")
        bastion_host = self._get_bastion_host(environment)
        bastion_username = self._fetch_config(environment, "bastion_username", "akeyless_mysql_bastion_username")
        public_key_file_path = self._fetch_config(
            environment,
            "public_key_file_path",
            "akeyless_mysql_bastion_public_key_file_path",
            default=default_public_key_file_path,
        )
        cert_issuer_name = self._fetch_config(
            environment, "cert_issuer_name", "akeyless_mysql_bastion_cert_issuer_name"
        )
        local_proxy_port = self._fetch_config(
            environment, "local_proxy_port", "akeyless_mysql_bastion_local_proxy_port", default=8888
        )
        database_host = self._fetch_config(environment, "database_host", "db_host")
        database_name = self._fetch_config(environment, "database_name", "db_database")

        # Create the SSH tunnel (yeah, it's obnoxious)
        self._create_tunnel(
            secrets,
            bastion_host,
            bastion_username,
            local_proxy_port,
            cert_issuer_name,
            public_key_file_path,
            database_host,
        )

        # and now we can fetch credentials
        credentials = secrets.get_dynamic_secret(producer_name)

        return {
            "username": credentials["user"],
            "password": credentials["password"],
            "host": "127.0.0.1",
            "database": database_name,
            "port": local_proxy_port,
        }

    def _get_bastion_host(self, environment):
        return self._fetch_config(environment, "bastion_host", "akeyless_mysql_bastion_host")

    def _fetch_config(self, environment, config_key_name, environment_key_name, default=None):
        if self.config[config_key_name]:
            return self.config[config_key_name]
        from_environment = environment.get(environment_key_name, silent=True)
        if from_environment:
            return from_environment
        if default is not None:
            return default
        raise ValueError(
            f"I was asked to connect to a database via an AKeyless dynamic producer through an SSH bastion"
            "with certificate auth, but I wasn't given a required configuration value: '{config_key_name}'."
            "This can be set in the call to "
            "`clearskies.backends.akeyless.mysql_connection_dynamic_producer_via_ssh_cert_bastion()` by providing the "
            "'{config_key_name}' argument, or by setting an environment variable named '{environment_key_name}'."
        )

    def _create_tunnel(
        self,
        secrets,
        bastion_host,
        bastion_username,
        local_proxy_port,
        cert_issuer_name,
        public_key_file_path,
        database_host,
    ):
        # first see if the socket is already open, since we don't close it.
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex(("127.0.0.1", local_proxy_port))
        if result == 0:
            sock.close()
            return

        if not os.path.isfile(public_key_file_path):
            raise ValueError(
                f"I was asked to connect to AKeyless SSH with the public key file in '{public_key_file_path}',"
                "but this file does not exist"
            )

        ssh_certificate = secrets.get_ssh_certificate(cert_issuer_name, bastion_username, public_key_file_path)

        # We need to write the certificate out to the standard location that SSH expects it so that SSH can find it.
        # I haven't found a good library for doing this in Python, so I'm relying on the ssh command
        home = str(Path.home())
        with open(f"{home}/.ssh/id_rsa-cert.pub", "w") as fp:
            fp.write(ssh_certificate)

        # and now we can do this thing.
        tunnel_command = [
            "ssh",
            "-o",
            "ConnectTimeout=2",
            "-N",
            "-L",
            f"{local_proxy_port}:{database_host}:3306",
            "-p",
            "22",
            f"{bastion_username}@{bastion_host}",
        ]
        subprocess.Popen(tunnel_command)
        connected = False
        attempts = 0
        while not connected and attempts < 6:
            attempts += 1
            time.sleep(0.5)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = sock.connect_ex(("127.0.0.1", local_proxy_port))
            if result == 0:
                connected = True
        if not connected:
            raise ValueError(
                "Failed to open SSH tunnel.  The following command was used: \n" + " ".join(tunnel_command)
            )



import datetime
import re


def camel_case_to_snake_case(string: str) -> str:
    """Convert a title/camel case string (MyString|myString) to snake case (my_string)."""
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)).lower()


def camel_case_to_title_case(string):
    return camel_case_to_words(string).title().replace(" ", "")


def camel_case_to_words(string):
    string = re.sub("(.)([A-Z][a-z]+)", r"\1 \2", string)
    string = re.sub("([a-z0-9])([A-Z])", r"\1 \2", string).lower()
    return string


def camel_case_to_nice(string):
    return camel_case_to_words(string).title()


def title_case_to_snake_case(string: str) -> str:
    """Convert a title case string (MyString) to snake case (my_string)."""
    return camel_case_to_snake_case(string)


def title_case_to_camel_case(string: str) -> str:
    if len(string) == 0:
        return string
    if len(string) == 1:
        return string.lower()
    return string[0].lower() + string[1:]


def title_case_to_nice(string: str) -> str:
    return camel_case_to_nice(string)


def snake_case_to_title_case(string: str) -> str:
    """
    Convert a snake case string (my_string) to title case (MyString).

    Note this is sometimes ambiguous.  Consider:

    TitleCase -> snake_case  -> TitleCase
    MyDbThing -> my_db_thing -> MyDbThing
    MyDBThing -> my_db_thing -> MyDbThing
    """
    words = string.lower().split("_")
    return "".join([x.title() for x in words])


def snake_case_to_camel_case(string: str) -> str:
    """
    Convert a snake case string (my_string) to camel case (myString).

    Note this is sometimes ambiguous.  Consider:

    camelCase -> snake_case  -> camelCase
    myDbThing -> my_db_thing -> myDbThing
    myDBThing -> my_db_thing -> myDbThing
    """
    words = string.lower().split("_")
    return words[0] + "".join([x.title() for x in words[1:]])


def snake_case_to_nice(string: str) -> str:
    return camel_case_to_nice(snake_case_to_camel_case(string))


casings = ["camelCase", "snake_case", "TitleCase"]
casing_swap_map = {
    "camelCase": {
        "camelCase": str,
        "snake_case": camel_case_to_snake_case,
        "TitleCase": camel_case_to_title_case,
    },
    "snake_case": {
        "camelCase": snake_case_to_camel_case,
        "snake_case": str,
        "TitleCase": snake_case_to_title_case,
    },
    "TitleCase": {
        "camelCase": title_case_to_camel_case,
        "snake_case": title_case_to_snake_case,
        "TitleCase": str,
    },
}


def swap_casing(string: str, from_casing: str, to_casing: str) -> str:
    if from_casing not in casings:
        raise ValueError(f"Invalid casing '{from_casing}'.  Must be one of '" + "', ".join(casings) + "'")
    if to_casing not in casings:
        raise ValueError(f"Invalid casing '{to_casing}'.  Must be one of '" + "', ".join(casings) + "'")
    return casing_swap_map[from_casing][to_casing](string)  # type: ignore


def make_plural(singular: str):
    if singular[-1] == "y":
        return singular[:-1] + "ies"
    if singular[-1] == "s":
        return singular + "es"
    return f"{singular}s"


def datetime_to_iso(value):
    if not isinstance(value, datetime.date) and not isinstance(value, datetime.datetime):
        return value
    return value.isoformat()



from .array import Array
from .base64 import Base64
from .boolean import Boolean
from .date import Date
from .datetime import Datetime
from .double import Double
from .enum import Enum
from .integer import Integer
from .long import Long
from .number import Number
from .object import Object
from .password import Password
from .schema import Schema
from .string import String

__all__ = [
    "Array",
    "Base64",
    "Boolean",
    "Date",
    "Datetime",
    "Double",
    "Enum",
    "Integer",
    "Long",
    "Number",
    "Object",
    "Password",
    "Schema",
    "String",
]



from clearskies.validators.after_column import AfterColumn
from clearskies.validators.before_column import BeforeColumn
from clearskies.validators.in_the_future import InTheFuture
from clearskies.validators.in_the_future_at_least import InTheFutureAtLeast
from clearskies.validators.in_the_future_at_most import InTheFutureAtMost
from clearskies.validators.in_the_past import InThePast
from clearskies.validators.in_the_past_at_least import InThePastAtLeast
from clearskies.validators.in_the_past_at_most import InThePastAtMost
from clearskies.validators.maximum_length import MaximumLength
from clearskies.validators.maximum_value import MaximumValue
from clearskies.validators.minimum_length import MinimumLength
from clearskies.validators.minimum_value import MinimumValue
from clearskies.validators.required import Required
from clearskies.validators.timedelta import Timedelta
from clearskies.validators.unique import Unique

__all__ = [
    "AfterColumn",
    "BeforeColumn",
    "InTheFuture",
    "InTheFutureAtLeast",
    "InTheFutureAtMost",
    "InThePast",
    "InThePastAtLeast",
    "InThePastAtMost",
    "MaximumLength",
    "MaximumValue",
    "MinimumLength",
    "MaximumLength",
    "Required",
    "Timedelta",
    "Unique",
]



from clearskies.endpoints.advanced_search import AdvancedSearch
from clearskies.endpoints.callable import Callable
from clearskies.endpoints.create import Create
from clearskies.endpoints.delete import Delete
from clearskies.endpoints.get import Get
from clearskies.endpoints.health_check import HealthCheck
from clearskies.endpoints.list import List
from clearskies.endpoints.restful_api import RestfulApi
from clearskies.endpoints.simple_search import SimpleSearch
from clearskies.endpoints.update import Update

__all__ = [
    "AdvancedSearch",
    "Callable",
    "Create",
    "Delete",
    "Get",
    "HealthCheck",
    "List",
    "RestfulApi",
    "SimpleSearch",
    "Update",
]



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

from clearskies.configs import config
from clearskies.functional import validations

if TYPE_CHECKING:
    from clearskies.model import Model, ModelClassReference


class ModelClass(config.Config):
    """A config that accepts a model class."""

    def __set__(self, instance, value: type[Model | ModelClassReference]):
        try:
            validations.is_model_class_or_reference(value, raise_error_message=True, strict=False)
        except TypeError as e:
            error_prefix = self._error_prefix(instance)
            raise TypeError(f"{error_prefix} {str(e)}")

        # reference or model class?
        instance._set_config(self, value)  # type: ignore

    def __get__(self, instance, parent) -> type[Model]:
        if not instance:
            return self  # type: ignore

        value = instance._get_config(self)
        if validations.is_model_class_reference(value):
            class_reference = value() if inspect.isclass(value) else value
            instance._set_config(self, class_reference.get_model_class())

        return instance._get_config(self)



from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable, Type

import clearskies.configs
import clearskies.exceptions
from clearskies import authentication, autodoc, typing
from clearskies.endpoints.get import Get
from clearskies.functional import routing, string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import SecurityHeader
    from clearskies.model import Model, Schema


class Delete(Get):
    """
    An endpoint that deletes a single record by id (or other unique column).

    This endpoint is intended to delete a single record.  You have to provide a model class and (optionally) the name
    of the column that it should use to lookup the matching record.  If you don't specifically tell it what column to
    use to lookup the record, it will assume that you want to use the id column of the model.  Finally, you must
    declare a route parameter with a matching column name: the delete endpoint will then fetch the desired record id
    out of the URL path.  The default request method is DELETE. Here's a simple example:

    ```python
    import clearskies


    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        username = clearskies.columns.String()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Delete(
            model_class=User,
            url="/{id}",
        ),
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": User,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob Brown", "username": "bobbrown"},
                        {"id": "1-2-3-5", "name": "Jane Doe", "username": "janedoe"},
                        {"id": "1-2-3-6", "name": "Greg", "username": "greg"},
                    ],
                },
            ],
        },
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080/1-2-3-4' -X DELETE | jq
    {
        "status": "success",
        "error": "",
        "data": {},
        "pagination": {},
        "input_errors": {}
    }
    ```
    """

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        model_class: type[Model],
        url: str,
        record_lookup_column_name: str | None = None,
        response_headers: list[str | Callable[..., list[str]]] = [],
        request_methods: list[str] = ["DELETE"],
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        where: typing.condition | list[typing.condition] = [],
        joins: typing.join | list[typing.join] = [],
        authentication: authentication.Authentication = authentication.Public(),
        authorization: authentication.Authorization = authentication.Authorization(),
    ):
        # see comment in clearskies.endpoints.Create.__init__
        self.request_methods = request_methods

        # we need to call the parent but don't have to pass along any of our kwargs.  They are all optional in our parent, and our parent class
        # just stores them in parameters, which we have already done.  However, the parent does do some extra initialization stuff that we need,
        # which is why we have to call the parent.
        super().__init__(model_class, url, [model_class.id_column_name])

    def handle(self, input_output: InputOutput) -> Any:
        model = self.fetch_model(input_output)
        model.delete()
        return self.success(input_output, {})

    def documentation(self) -> list[autodoc.request.Request]:
        output_autodoc = (autodoc.schema.Object(self.auto_case_internal_column_name("data"), children={}),)

        authentication = self.authentication
        standard_error_responses = [self.documentation_input_error_response()]
        if not getattr(authentication, "is_public", False):
            standard_error_responses.append(self.documentation_access_denied_response())
            if getattr(authentication, "can_authorize", False):
                standard_error_responses.append(self.documentation_unauthorized_response())

        return [
            autodoc.request.Request(
                self.description,
                [
                    self.documentation_success_response(
                        output_autodoc,  # type: ignore
                        description=self.description,
                    ),
                    *standard_error_responses,
                    self.documentation_generic_error_response(),
                ],
                relative_path=self.url,
                request_methods=self.request_methods,
                parameters=[
                    *self.documentation_routing_parameters(),
                ],
                root_properties={
                    "security": self.documentation_request_security(),
                },
            ),
        ]

    def documentation_models(self) -> dict[str, autodoc.schema.Schema]:
        return {}



import json
from abc import ABC, abstractmethod
from typing import Any
from urllib.parse import parse_qs

import clearskies.configurable
import clearskies.typing
from clearskies.configs import AnyDict, StringDict
from clearskies.exceptions import ClientError

from .headers import Headers


class InputOutput(ABC, clearskies.configurable.Configurable):
    """Manage the request and response to the client."""

    response_headers: Headers = None  # type: ignore
    request_headers: Headers = None  # type: ignore
    query_parameters = clearskies.configs.AnyDict(default={})
    routing_data = clearskies.configs.StringDict(default={})
    authorization_data = clearskies.configs.AnyDict(default={})

    _body_as_json: dict[str, Any] | list[Any] | None = {}
    _body_loaded_as_json = False

    def __init__(self):
        self.response_headers = Headers()
        self.request_headers = Headers(self.get_request_headers())
        self.query_parameters = {key: val[0] for (key, val) in parse_qs(self.get_query_string()).items()}
        self.authorization_data = {}
        self.routing_data = {}
        self.finalize_and_validate_configuration()

    @abstractmethod
    def respond(self, body: clearskies.typing.response, status_code: int = 200) -> Any:
        """
        Pass along a response to the client.

        Accepts a string, bytes, dictionary, or list.  If a content type has not been set then it will automatically
        be set to application/json
        """
        pass

    @abstractmethod
    def get_body(self) -> str:
        """Return the raw body set by the client."""
        pass

    @abstractmethod
    def has_body(self) -> bool:
        """Whether or not the request included a body."""
        pass

    @property
    def request_data(self) -> dict[str, Any] | list[Any] | None:
        """Return the data from the request body, assuming it is JSON."""
        if not self._body_loaded_as_json:
            self._body_loaded_as_json = True
            if not self.has_body():
                self._body_as_json = None
            else:
                try:
                    self._body_as_json = json.loads(self.get_body())
                except json.JSONDecodeError:
                    self._body_as_json = None
        return self._body_as_json

    @abstractmethod
    def get_request_method(self) -> str:
        """Return the request method set by the client."""
        pass

    @abstractmethod
    def get_script_name(self) -> str:
        """Return the script name, e.g. the path requested."""
        pass

    @abstractmethod
    def get_path_info(self) -> str:
        """Return the path info for the request."""
        pass

    @abstractmethod
    def get_query_string(self) -> str:
        """Return the full query string for the request (everything after the first question mark in the document URL)."""
        pass

    @abstractmethod
    def get_client_ip(self):
        pass

    @abstractmethod
    def get_request_headers(self) -> dict[str, str]:
        pass

    def get_full_path(self) -> str:
        """Return the full path requested by the client."""
        path_info = self.get_path_info()
        script_name = self.get_script_name()
        if not path_info or path_info[0] != "/":
            path_info = f"/{path_info}"
        return f"{path_info}{script_name}".replace("//", "/")

    def context_specifics(self):
        return {}

    def get_context_for_callables(self) -> dict[str, Any]:
        """
        Return a dictionary with various important parts of the request that are passed along to user-defined functions.

        It's common to make various aspects of an incoming request available to user-defined functions that are
        attached to clearskies hooks everywhere.  This function centralizes the definition of what aspects of
        the reequest shouuld be passed along to callables in this case.  When this is in use it typically
        looks like this:

        di.call_function(some_function, **input_output.get_context_for_callables())

        And this function returns a dictionary with the following values:

        | Key                | Type                             | Ref                             | Value                                                                           |
        |--------------------|----------------------------------|---------------------------------|---------------------------------------------------------------------------------|
        | routing_data       | dict[str, str]                   | input_output.routing_data       | A dictionary of data extracted from URL path parameters.                        |
        | authorization_data | dict[str, Any]                   | input_output.authorization_data | A dictionary containing the authorization data set by the authentication method |
        | request_data       | dict[str, Any] | None            | input_output.request_data       | The data sent along with the request (assuming a JSON request body)             |
        | query_parameters   | dict[str, Any]                   | input_output.query_parameters   | The query parameters                                                            |
        | request_headers    | clearskies.input_outputs.Headers | input_output.request_headers    | The request headers sent by the client                                          |
        | **routing_data     | string                           | **input_output.routing_data     | The routing data is unpacked so keys can be fetched directly                    |
        """
        return {
            **self.routing_data,
            **{
                "routing_data": self.routing_data,
                "authorization_data": self.authorization_data,
                "request_data": self.request_data,
                "request_headers": self.request_headers,
                "query_parameters": self.query_parameters,
            },
        }



from .schema import Schema


class Object(Schema):
    children = None
    model_name = None
    _type = "object"
    _format = ""

    def __init__(self, name, children, value=None, example=None, model_name=None):
        super().__init__(name, example=example, value=value)
        self.children = children
        self.model_name = model_name
